"""Test stepping over watchpoints."""



import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestStepOverWatchpoint(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    @expectedFailureAll(
        oslist=["freebsd", "linux"],
        archs=[
            'aarch64',
            'arm'],
        bugnumber="llvm.org/pr26031")
    # Read-write watchpoints not supported on SystemZ
    @expectedFailureAll(archs=['s390x'])
    @expectedFailureAll(
        oslist=["ios", "watchos", "tvos", "bridgeos", "macosx"],
        archs=['aarch64', 'arm'],
        bugnumber="<rdar://problem/34027183>")
    @add_test_categories(["basic_process"])
    def test(self):
        """Test stepping over watchpoints."""
        self.build()
        target = self.createTestTarget()

        lldbutil.run_break_set_by_symbol(self, 'main')

        process = target.LaunchSimple(None, None,
                                      self.get_process_working_directory())
        self.assertTrue(process.IsValid(), PROCESS_IS_VALID)
        self.assertState(process.GetState(), lldb.eStateStopped,
                         PROCESS_STOPPED)

        thread = lldbutil.get_stopped_thread(process,
                                             lldb.eStopReasonBreakpoint)
        self.assertTrue(thread.IsValid(), "Failed to get thread.")

        frame = thread.GetFrameAtIndex(0)
        self.assertTrue(frame.IsValid(), "Failed to get frame.")

        read_value = frame.FindValue('g_watch_me_read',
                                     lldb.eValueTypeVariableGlobal)
        self.assertTrue(read_value.IsValid(), "Failed to find read value.")

        error = lldb.SBError()

        # resolve_location=True, read=True, write=False
        read_watchpoint = read_value.Watch(True, True, False, error)
        self.assertSuccess(error, "Error while setting watchpoint")
        self.assertTrue(read_watchpoint, "Failed to set read watchpoint.")

        thread.StepOver()
        self.assertEquals(thread.GetStopReason(), lldb.eStopReasonWatchpoint,
                        STOPPED_DUE_TO_WATCHPOINT)
        self.assertEquals(thread.GetStopDescription(20), 'watchpoint 1')

        process.Continue()
        self.assertState(process.GetState(), lldb.eStateStopped,
                         PROCESS_STOPPED)
        self.assertEquals(thread.GetStopDescription(20), 'step over')

        self.step_inst_for_watchpoint(1)

        write_value = frame.FindValue('g_watch_me_write',
                                      lldb.eValueTypeVariableGlobal)
        self.assertTrue(write_value, "Failed to find write value.")

        # Most of the MIPS boards provide only one H/W watchpoints, and S/W
        # watchpoints are not supported yet
        arch = self.getArchitecture()
        if re.match("^mips", arch) or re.match("powerpc64le", arch):
            self.runCmd("watchpoint delete 1")

        # resolve_location=True, read=False, write=True
        write_watchpoint = write_value.Watch(True, False, True, error)
        self.assertTrue(write_watchpoint, "Failed to set write watchpoint.")
        self.assertSuccess(error, "Error while setting watchpoint")

        thread.StepOver()
        self.assertEquals(thread.GetStopReason(), lldb.eStopReasonWatchpoint,
                        STOPPED_DUE_TO_WATCHPOINT)
        self.assertEquals(thread.GetStopDescription(20), 'watchpoint 2')

        process.Continue()
        self.assertState(process.GetState(), lldb.eStateStopped,
                         PROCESS_STOPPED)
        self.assertEquals(thread.GetStopDescription(20), 'step over')

        self.step_inst_for_watchpoint(2)

    def step_inst_for_watchpoint(self, wp_id):
        watchpoint_hit = False
        current_line = self.frame().GetLineEntry().GetLine()
        while self.frame().GetLineEntry().GetLine() == current_line:
            self.thread().StepInstruction(False)  # step_over=False
            stop_reason = self.thread().GetStopReason()
            if stop_reason == lldb.eStopReasonWatchpoint:
                self.assertFalse(watchpoint_hit, "Watchpoint already hit.")
                expected_stop_desc = "watchpoint %d" % wp_id
                actual_stop_desc = self.thread().GetStopDescription(20)
                self.assertEquals(actual_stop_desc, expected_stop_desc,
                                "Watchpoint ID didn't match.")
                watchpoint_hit = True
            else:
                self.assertEquals(stop_reason, lldb.eStopReasonPlanComplete,
                                STOPPED_DUE_TO_STEP_IN)
        self.assertTrue(watchpoint_hit, "Watchpoint never hit.")



"""
Test example snippets from the lldb 'help expression' output.
"""



import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class Radar9673644TestCase(TestBase):

    def setUp(self):
        # Call super's setUp().
        TestBase.setUp(self)
        # Find the line number to break inside main().
        self.main_source = "main.c"
        self.line = line_number(self.main_source, '// Set breakpoint here.')

    def test_expr_commands(self):
        """The following expression commands should just work."""
        self.build()

        self.runCmd("file " + self.getBuildArtifact("a.out"), CURRENT_EXECUTABLE_SET)

        lldbutil.run_break_set_by_file_and_line(
            self,
            self.main_source,
            self.line,
            num_expected_locations=1,
            loc_exact=True)

        self.runCmd("run", RUN_SUCCEEDED)

        # rdar://problem/9673664 lldb expression evaluation problem

        self.expect('expr char str[] = "foo"; str[0]',
                    substrs=["'f'"])
        # runCmd: expr char c[] = "foo"; c[0]
        # output: (char) $0 = 'f'



"""
Test the output of `frame diagnose` for a subexpression of a complicated expression
"""


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestDiagnoseDereferenceArgument(TestBase):

    @skipUnlessDarwin
    @skipIf(archs=no_match(['x86_64'])) # <rdar://problem/33842388> frame diagnose doesn't work for armv7 or arm64
    def test_diagnose_dereference_argument(self):
        TestBase.setUp(self)
        self.build()
        exe = self.getBuildArtifact("a.out")
        self.runCmd("file " + exe, CURRENT_EXECUTABLE_SET)
        self.runCmd("run", RUN_SUCCEEDED)
        self.expect("thread list", "Thread should be stopped",
                    substrs=['stopped'])
        self.expect(
            "frame diagnose",
            "Crash diagnosis was accurate",
            substrs=["f->b->d"])



# DExTer : Debugging Experience Tester
# ~~~~~~   ~         ~~         ~   ~~
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ctypes import *
from functools import partial

from .utils import *
from .breakpoint import *

class DEBUG_STACK_FRAME_EX(Structure):
  _fields_ = [
      ("InstructionOffset", c_ulonglong),
      ("ReturnOffset", c_ulonglong),
      ("FrameOffset", c_ulonglong),
      ("StackOffset", c_ulonglong),
      ("FuncTableEntry", c_ulonglong),
      ("Params", c_ulonglong * 4),
      ("Reserved", c_ulonglong * 6),
      ("Virtual", c_bool),
      ("FrameNumber", c_ulong),
      ("InlineFrameContext", c_ulong),
      ("Reserved1", c_ulong)
    ]
PDEBUG_STACK_FRAME_EX = POINTER(DEBUG_STACK_FRAME_EX)

class DEBUG_VALUE_U(Union):
  _fields_ = [
      ("I8", c_byte),
      ("I16", c_short),
      ("I32", c_int),
      ("I64", c_long),
      ("F32", c_float),
      ("F64", c_double),
      ("RawBytes", c_ubyte * 24) # Force length to 24b.
    ]

class DEBUG_VALUE(Structure):
  _fields_ = [
      ("U", DEBUG_VALUE_U),
      ("TailOfRawBytes", c_ulong),
      ("Type", c_ulong)
    ]
PDEBUG_VALUE = POINTER(DEBUG_VALUE)

class DebugValueType(IntEnum):
  DEBUG_VALUE_INVALID      = 0
  DEBUG_VALUE_INT8         = 1
  DEBUG_VALUE_INT16        = 2
  DEBUG_VALUE_INT32        = 3
  DEBUG_VALUE_INT64        = 4
  DEBUG_VALUE_FLOAT32      = 5
  DEBUG_VALUE_FLOAT64      = 6
  DEBUG_VALUE_FLOAT80      = 7
  DEBUG_VALUE_FLOAT82      = 8
  DEBUG_VALUE_FLOAT128     = 9
  DEBUG_VALUE_VECTOR64     = 10
  DEBUG_VALUE_VECTOR128    = 11
  DEBUG_VALUE_TYPES        = 12

# UUID for DebugControl7 interface.
DebugControl7IID = IID(0xb86fb3b1, 0x80d4, 0x475b, IID_Data4_Type(0xae, 0xa3, 0xcf, 0x06, 0x53, 0x9c, 0xf6, 0x3a))

class IDebugControl7(Structure):
  pass

class IDebugControl7Vtbl(Structure):
  wrp = partial(WINFUNCTYPE, c_long, POINTER(IDebugControl7))
  idc_getnumbereventfilters = wrp(c_ulong_p, c_ulong_p, c_ulong_p)
  idc_setexceptionfiltersecondcommand = wrp(c_ulong, c_char_p)
  idc_waitforevent = wrp(c_long, c_long)
  idc_execute = wrp(c_long, c_char_p, c_long)
  idc_setexpressionsyntax = wrp(c_ulong)
  idc_addbreakpoint2 = wrp(c_ulong, c_ulong, POINTER(POINTER(DebugBreakpoint2)))
  idc_setexecutionstatus = wrp(c_ulong)
  idc_getexecutionstatus = wrp(c_ulong_p)
  idc_getstacktraceex = wrp(c_ulonglong, c_ulonglong, c_ulonglong, PDEBUG_STACK_FRAME_EX, c_ulong, c_ulong_p)
  idc_evaluate = wrp(c_char_p, c_ulong, PDEBUG_VALUE, c_ulong_p)
  idc_setengineoptions = wrp(c_ulong)
  _fields_ = [
      ("QueryInterface", c_void_p),
      ("AddRef", c_void_p),
      ("Release", c_void_p),
      ("GetInterrupt", c_void_p),
      ("SetInterrupt", c_void_p),
      ("GetInterruptTimeout", c_void_p),
      ("SetInterruptTimeout", c_void_p),
      ("GetLogFile", c_void_p),
      ("OpenLogFile", c_void_p),
      ("CloseLogFile", c_void_p),
      ("GetLogMask", c_void_p),
      ("SetLogMask", c_void_p),
      ("Input", c_void_p),
      ("ReturnInput", c_void_p),
      ("Output", c_void_p),
      ("OutputVaList", c_void_p),
      ("ControlledOutput", c_void_p),
      ("ControlledOutputVaList", c_void_p),
      ("OutputPrompt", c_void_p),
      ("OutputPromptVaList", c_void_p),
      ("GetPromptText", c_void_p),
      ("OutputCurrentState", c_void_p),
      ("OutputVersionInformation", c_void_p),
      ("GetNotifyEventHandle", c_void_p),
      ("SetNotifyEventHandle", c_void_p),
      ("Assemble", c_void_p),
      ("Disassemble", c_void_p),
      ("GetDisassembleEffectiveOffset", c_void_p),
      ("OutputDisassembly", c_void_p),
      ("OutputDisassemblyLines", c_void_p),
      ("GetNearInstruction", c_void_p),
      ("GetStackTrace", c_void_p),
      ("GetReturnOffset", c_void_p),
      ("OutputStackTrace", c_void_p),
      ("GetDebuggeeType", c_void_p),
      ("GetActualProcessorType", c_void_p),
      ("GetExecutingProcessorType", c_void_p),
      ("GetNumberPossibleExecutingProcessorTypes", c_void_p),
      ("GetPossibleExecutingProcessorTypes", c_void_p),
      ("GetNumberProcessors", c_void_p),
      ("GetSystemVersion", c_void_p),
      ("GetPageSize", c_void_p),
      ("IsPointer64Bit", c_void_p),
      ("ReadBugCheckData", c_void_p),
      ("GetNumberSupportedProcessorTypes", c_void_p),
      ("GetSupportedProcessorTypes", c_void_p),
      ("GetProcessorTypeNames", c_void_p),
      ("GetEffectiveProcessorType", c_void_p),
      ("SetEffectiveProcessorType", c_void_p),
      ("GetExecutionStatus", idc_getexecutionstatus),
      ("SetExecutionStatus", idc_setexecutionstatus),
      ("GetCodeLevel", c_void_p),
      ("SetCodeLevel", c_void_p),
      ("GetEngineOptions", c_void_p),
      ("AddEngineOptions", c_void_p),
      ("RemoveEngineOptions", c_void_p),
      ("SetEngineOptions", idc_setengineoptions),
      ("GetSystemErrorControl", c_void_p),
      ("SetSystemErrorControl", c_void_p),
      ("GetTextMacro", c_void_p),
      ("SetTextMacro", c_void_p),
      ("GetRadix", c_void_p),
      ("SetRadix", c_void_p),
      ("Evaluate", idc_evaluate),
      ("CoerceValue", c_void_p),
      ("CoerceValues", c_void_p),
      ("Execute", idc_execute),
      ("ExecuteCommandFile", c_void_p),
      ("GetNumberBreakpoints", c_void_p),
      ("GetBreakpointByIndex", c_void_p),
      ("GetBreakpointById", c_void_p),
      ("GetBreakpointParameters", c_void_p),
      ("AddBreakpoint", c_void_p),
      ("RemoveBreakpoint", c_void_p),
      ("AddExtension", c_void_p),
      ("RemoveExtension", c_void_p),
      ("GetExtensionByPath", c_void_p),
      ("CallExtension", c_void_p),
      ("GetExtensionFunction", c_void_p),
      ("GetWindbgExtensionApis32", c_void_p),
      ("GetWindbgExtensionApis64", c_void_p),
      ("GetNumberEventFilters", idc_getnumbereventfilters),
      ("GetEventFilterText", c_void_p),
      ("GetEventFilterCommand", c_void_p),
      ("SetEventFilterCommand", c_void_p),
      ("GetSpecificFilterParameters", c_void_p),
      ("SetSpecificFilterParameters", c_void_p),
      ("GetSpecificFilterArgument", c_void_p),
      ("SetSpecificFilterArgument", c_void_p),
      ("GetExceptionFilterParameters", c_void_p),
      ("SetExceptionFilterParameters", c_void_p),
      ("GetExceptionFilterSecondCommand", c_void_p),
      ("SetExceptionFilterSecondCommand", idc_setexceptionfiltersecondcommand),
      ("WaitForEvent", idc_waitforevent),
      ("GetLastEventInformation", c_void_p),
      ("GetCurrentTimeDate", c_void_p),
      ("GetCurrentSystemUpTime", c_void_p),
      ("GetDumpFormatFlags", c_void_p),
      ("GetNumberTextReplacements", c_void_p),
      ("GetTextReplacement", c_void_p),
      ("SetTextReplacement", c_void_p),
      ("RemoveTextReplacements", c_void_p),
      ("OutputTextReplacements", c_void_p),
      ("GetAssemblyOptions", c_void_p),
      ("AddAssemblyOptions", c_void_p),
      ("RemoveAssemblyOptions", c_void_p),
      ("SetAssemblyOptions", c_void_p),
      ("GetExpressionSyntax", c_void_p),
      ("SetExpressionSyntax", idc_setexpressionsyntax),
      ("SetExpressionSyntaxByName", c_void_p),
      ("GetNumberExpressionSyntaxes", c_void_p),
      ("GetExpressionSyntaxNames", c_void_p),
      ("GetNumberEvents", c_void_p),
      ("GetEventIndexDescription", c_void_p),
      ("GetCurrentEventIndex", c_void_p),
      ("SetNextEventIndex", c_void_p),
      ("GetLogFileWide", c_void_p),
      ("OpenLogFileWide", c_void_p),
      ("InputWide", c_void_p),
      ("ReturnInputWide", c_void_p),
      ("OutputWide", c_void_p),
      ("OutputVaListWide", c_void_p),
      ("ControlledOutputWide", c_void_p),
      ("ControlledOutputVaListWide", c_void_p),
      ("OutputPromptWide", c_void_p),
      ("OutputPromptVaListWide", c_void_p),
      ("GetPromptTextWide", c_void_p),
      ("AssembleWide", c_void_p),
      ("DisassembleWide", c_void_p),
      ("GetProcessrTypeNamesWide", c_void_p),
      ("GetTextMacroWide", c_void_p),
      ("SetTextMacroWide", c_void_p),
      ("EvaluateWide", c_void_p),
      ("ExecuteWide", c_void_p),
      ("ExecuteCommandFileWide", c_void_p),
      ("GetBreakpointByIndex2", c_void_p),
      ("GetBreakpointById2", c_void_p),
      ("AddBreakpoint2", idc_addbreakpoint2),
      ("RemoveBreakpoint2", c_void_p),
      ("AddExtensionWide", c_void_p),
      ("GetExtensionByPathWide", c_void_p),
      ("CallExtensionWide", c_void_p),
      ("GetExtensionFunctionWide", c_void_p),
      ("GetEventFilterTextWide", c_void_p),
      ("GetEventfilterCommandWide", c_void_p),
      ("SetEventFilterCommandWide", c_void_p),
      ("GetSpecificFilterArgumentWide", c_void_p),
      ("SetSpecificFilterArgumentWide", c_void_p),
      ("GetExceptionFilterSecondCommandWide", c_void_p),
      ("SetExceptionFilterSecondCommandWider", c_void_p),
      ("GetLastEventInformationWide", c_void_p),
      ("GetTextReplacementWide", c_void_p),
      ("SetTextReplacementWide", c_void_p),
      ("SetExpressionSyntaxByNameWide", c_void_p),
      ("GetExpressionSyntaxNamesWide", c_void_p),
      ("GetEventIndexDescriptionWide", c_void_p),
      ("GetLogFile2", c_void_p),
      ("OpenLogFile2", c_void_p),
      ("GetLogFile2Wide", c_void_p),
      ("OpenLogFile2Wide", c_void_p),
      ("GetSystemVersionValues", c_void_p),
      ("GetSystemVersionString", c_void_p),
      ("GetSystemVersionStringWide", c_void_p),
      ("GetContextStackTrace", c_void_p),
      ("OutputContextStackTrace", c_void_p),
      ("GetStoredEventInformation", c_void_p),
      ("GetManagedStatus", c_void_p),
      ("GetManagedStatusWide", c_void_p),
      ("ResetManagedStatus", c_void_p),
      ("GetStackTraceEx", idc_getstacktraceex),
      ("OutputStackTraceEx", c_void_p),
      ("GetContextStackTraceEx", c_void_p),
      ("OutputContextStackTraceEx", c_void_p),
      ("GetBreakpointByGuid", c_void_p),
      ("GetExecutionStatusEx", c_void_p),
      ("GetSynchronizationStatus", c_void_p),
      ("GetDebuggeeType2", c_void_p)
    ]

IDebugControl7._fields_ = [("lpVtbl", POINTER(IDebugControl7Vtbl))]

class DebugStatus(IntEnum):
  DEBUG_STATUS_NO_CHANGE =            0
  DEBUG_STATUS_GO =                   1
  DEBUG_STATUS_GO_HANDLED =           2
  DEBUG_STATUS_GO_NOT_HANDLED =       3
  DEBUG_STATUS_STEP_OVER =            4
  DEBUG_STATUS_STEP_INTO =            5
  DEBUG_STATUS_BREAK =                6
  DEBUG_STATUS_NO_DEBUGGEE =          7
  DEBUG_STATUS_STEP_BRANCH =          8
  DEBUG_STATUS_IGNORE_EVENT =         9
  DEBUG_STATUS_RESTART_REQUESTED =   10
  DEBUG_STATUS_REVERSE_GO =          11
  DEBUG_STATUS_REVERSE_STEP_BRANCH = 12
  DEBUG_STATUS_REVERSE_STEP_OVER =   13
  DEBUG_STATUS_REVERSE_STEP_INTO =   14
  DEBUG_STATUS_OUT_OF_SYNC =         15
  DEBUG_STATUS_WAIT_INPUT =          16
  DEBUG_STATUS_TIMEOUT =             17

class DebugSyntax(IntEnum):
  DEBUG_EXPR_MASM = 0
  DEBUG_EXPR_CPLUSPLUS = 1

class Control(object):
  def __init__(self, control):
    self.ptr = control
    self.control = control.contents
    self.vt = self.control.lpVtbl.contents
    # Keep a handy ulong for passing into C methods.
    self.ulong = c_ulong()

  def GetExecutionStatus(self, doprint=False):
    ret = self.vt.GetExecutionStatus(self.control, byref(self.ulong))
    aborter(ret, "GetExecutionStatus")
    status = DebugStatus(self.ulong.value)
    if doprint:
      print("Execution status: {}".format(status))
    return status

  def SetExecutionStatus(self, status):
    assert isinstance(status, DebugStatus)
    res = self.vt.SetExecutionStatus(self.control, status.value)
    aborter(res, "SetExecutionStatus")

  def WaitForEvent(self, timeout=100):
    # No flags are taken by WaitForEvent, hence 0
    ret = self.vt.WaitForEvent(self.control, 0, timeout)
    aborter(ret, "WaitforEvent", ignore=[S_FALSE])
    return ret

  def GetNumberEventFilters(self):
    specific_events = c_ulong()
    specific_exceptions = c_ulong()
    arbitrary_exceptions = c_ulong()
    res = self.vt.GetNumberEventFilters(self.control, byref(specific_events),
                                    byref(specific_exceptions),
                                    byref(arbitrary_exceptions))
    aborter(res, "GetNumberEventFilters")
    return (specific_events.value, specific_exceptions.value,
            arbitrary_exceptions.value)

  def SetExceptionFilterSecondCommand(self, index, command):
    buf = create_string_buffer(command.encode('ascii'))
    res = self.vt.SetExceptionFilterSecondCommand(self.control, index, buf)
    aborter(res, "SetExceptionFilterSecondCommand")
    return

  def AddBreakpoint2(self, offset=None, enabled=None):
    breakpoint = POINTER(DebugBreakpoint2)()
    res = self.vt.AddBreakpoint2(self.control, BreakpointTypes.DEBUG_BREAKPOINT_CODE, DEBUG_ANY_ID, byref(breakpoint))
    aborter(res, "Add breakpoint 2")
    bp = Breakpoint(breakpoint)

    if offset is not None:
      bp.SetOffset(offset)
    if enabled is not None and enabled:
      bp.SetFlags(BreakpointFlags.DEBUG_BREAKPOINT_ENABLED)

    return bp

  def RemoveBreakpoint(self, bp):
    res = self.vt.RemoveBreakpoint2(self.control, bp.breakpoint)
    aborter(res, "RemoveBreakpoint2")
    bp.die()

  def GetStackTraceEx(self):
    # XXX -- I can't find a way to query for how many stack frames there _are_
    # in  advance. Guess 128 for now.
    num_frames_buffer = 128

    frames = (DEBUG_STACK_FRAME_EX * num_frames_buffer)()
    numframes = c_ulong()

    # First three args are frame/stack/IP offsets -- leave them as zero to
    # default to the current instruction.
    res = self.vt.GetStackTraceEx(self.control, 0, 0, 0, frames, num_frames_buffer, byref(numframes))
    aborter(res, "GetStackTraceEx")
    return frames, numframes.value

  def Execute(self, command):
    # First zero is DEBUG_OUTCTL_*, which we leave as a default, second
    # zero is DEBUG_EXECUTE_* flags, of which we set none.
    res = self.vt.Execute(self.control, 0, command.encode('ascii'), 0)
    aborter(res, "Client execute")

  def SetExpressionSyntax(self, cpp=True):
    if cpp:
      syntax = DebugSyntax.DEBUG_EXPR_CPLUSPLUS
    else:
      syntax = DebugSyntax.DEBUG_EXPR_MASM

    res = self.vt.SetExpressionSyntax(self.control, syntax)
    aborter(res, "SetExpressionSyntax")

  def Evaluate(self, expr):
    ptr = DEBUG_VALUE()
    res = self.vt.Evaluate(self.control, expr.encode("ascii"), DebugValueType.DEBUG_VALUE_INVALID, byref(ptr), None)
    aborter(res, "Evaluate", ignore=[E_INTERNALEXCEPTION, E_FAIL])
    if res != 0:
      return None

    val_type = DebugValueType(ptr.Type)

    # Here's a map from debug value types to fields. Unclear what happens
    # with unsigned values, as DbgEng doesn't present any unsigned fields.

    extract_map = {
      DebugValueType.DEBUG_VALUE_INT8    : ("I8", "char"),
      DebugValueType.DEBUG_VALUE_INT16   : ("I16", "short"),
      DebugValueType.DEBUG_VALUE_INT32   : ("I32", "int"),
      DebugValueType.DEBUG_VALUE_INT64   : ("I64", "long"),
      DebugValueType.DEBUG_VALUE_FLOAT32 : ("F32", "float"),
      DebugValueType.DEBUG_VALUE_FLOAT64 : ("F64", "double")
    } # And everything else is invalid.

    if val_type not in extract_map:
      raise Exception("Unexpected debug value type {} when evalutaing".format(val_type))

    # Also produce a type name...

    return getattr(ptr.U, extract_map[val_type][0]), extract_map[val_type][1]

  def SetEngineOptions(self, opt):
    res = self.vt.SetEngineOptions(self.control, opt)
    aborter(res, "SetEngineOptions")
    return



import lldb


class fooSynthProvider:

    def __init__(self, valobj, dict):
        self.valobj = valobj
        self.int_type = valobj.GetType().GetBasicType(lldb.eBasicTypeInt)

    def num_children(self):
        return 3

    def get_child_at_index(self, index):
        if index == 0:
            child = self.valobj.GetChildMemberWithName('a')
        if index == 1:
            child = self.valobj.CreateChildAtOffset('fake_a', 1, self.int_type)
        if index == 2:
            child = self.valobj.GetChildMemberWithName('r')
        return child

    def get_child_index(self, name):
        if name == 'a':
            return 0
        if name == 'fake_a':
            return 1
        return 2

    def update(self):
        return True


class wrapfooSynthProvider:

    def __init__(self, valobj, dict):
        self.valobj = valobj

    def num_children(self):
        return 1

    def get_child_at_index(self, index):
        if index == 0:
            return self.valobj.GetChildMemberWithName('ptr')
        if index == 1:
            return self.valobj.GetChildMemberWithName('ptr').Dereference()
        return None

    def get_child_index(self, name):
        if name == 'ptr':
            return 0
        if name == '$$dereference$$':
            return 1
        return -1

    def update(self):
        return True



from __future__ import absolute_import

from lldbsuite.test import lldbinline

lldbinline.MakeInlineTest(
    __file__, globals(), [
        lldbinline.expectedFailureAll(
            oslist=["windows"], bugnumber="llvm.org/pr24663")])



# DExTer : Debugging Experience Tester
# ~~~~~~   ~         ~~         ~   ~~
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Commmand sets the path for all following commands to 'declared_file'.
"""

from pathlib import PurePath

from dex.command.CommandBase import CommandBase


class DexDeclareFile(CommandBase):
    def __init__(self, declared_file):

        if not isinstance(declared_file, str):
            raise TypeError('invalid argument type')

        # Use PurePath to create a cannonical platform path.
        # TODO: keep paths as PurePath objects for 'longer'
        self.declared_file = str(PurePath(declared_file))
        super(DexDeclareFile, self).__init__()

    @staticmethod
    def get_name():
        return __class__.__name__

    def eval(self):
        return self.declared_file



# DExTer : Debugging Experience Tester
# ~~~~~~   ~         ~~         ~   ~~
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Set of data classes for representing the complete debug program state at a
fixed point in execution.
"""

import os

from collections import OrderedDict
from pathlib import PurePath
from typing import List

class SourceLocation:
    def __init__(self, path: str = None, lineno: int = None, column: int = None):
        if path:
            path = os.path.normcase(path)
        self.path = path
        self.lineno = lineno
        self.column = column

    def __str__(self):
        return '{}({}:{})'.format(self.path, self.lineno, self.column)

    def match(self, other) -> bool:
        """Returns true iff all the properties that appear in `self` have the
        same value in `other`, but not necessarily vice versa.
        """
        if not other or not isinstance(other, SourceLocation):
            return False

        if self.path and (other.path is None or (PurePath(self.path) != PurePath(other.path))):
            return False

        if self.lineno and (self.lineno != other.lineno):
            return False

        if self.column and (self.column != other.column):
            return False

        return True


class StackFrame:
    def __init__(self,
                 function: str = None,
                 is_inlined: bool = None,
                 location: SourceLocation = None,
                 watches: OrderedDict = None):
        if watches is None:
            watches = {}

        self.function = function
        self.is_inlined = is_inlined
        self.location = location
        self.watches = watches

    def __str__(self):
        return '{}{}: {} | {}'.format(
            self.function,
            ' (inlined)' if self.is_inlined else '',
            self.location,
            {k: str(self.watches[k]) for k in self.watches})

    def match(self, other) -> bool:
        """Returns true iff all the properties that appear in `self` have the
        same value in `other`, but not necessarily vice versa.
        """
        if not other or not isinstance(other, StackFrame):
            return False

        if self.location and not self.location.match(other.location):
            return False

        if self.watches:
            for name in iter(self.watches):
                try:
                    if isinstance(self.watches[name], dict):
                        for attr in iter(self.watches[name]):
                            if (getattr(other.watches[name], attr, None) !=
                                    self.watches[name][attr]):
                                return False
                    else:
                        if other.watches[name].value != self.watches[name]:
                            return False
                except KeyError:
                    return False

        return True

class ProgramState:
    def __init__(self, frames: List[StackFrame] = None):
        self.frames = frames

    def __str__(self):
        return '\n'.join(map(
            lambda enum: 'Frame {}: {}'.format(enum[0], enum[1]),
            enumerate(self.frames)))

    def match(self, other) -> bool:
        """Returns true iff all the properties that appear in `self` have the
        same value in `other`, but not necessarily vice versa.
        """
        if not other or not isinstance(other, ProgramState):
            return False

        if self.frames:
            for idx, frame in enumerate(self.frames):
                try:
                    if not frame.match(other.frames[idx]):
                        return False
                except (IndexError, KeyError):
                    return False

        return True



from __future__ import print_function
import lldb
import os
from lldbsuite.test.lldbtest import *
from lldbsuite.test.decorators import *


@skipIfRemote
class TestQemuAPI(TestBase):
    NO_DEBUG_INFO_TESTCASE = True

    def test_file_api(self):
        qemu = lldb.SBPlatform("qemu-user")
        host = lldb.SBPlatform.GetHostPlatform()

        target = self.getBuildArtifact("target.c")
        main_c = lldb.SBFileSpec(self.getSourcePath("main.c"))

        self.assertSuccess(qemu.Put(main_c, lldb.SBFileSpec(target)))
        self.assertTrue(os.path.exists(target))
        self.assertEqual(qemu.GetFilePermissions(target),
                host.GetFilePermissions(target))

        self.assertSuccess(qemu.MakeDirectory(
            self.getBuildArtifact("target_dir")))
        self.assertTrue(os.path.isdir(self.getBuildArtifact("target_dir")))



# Originally imported via:
#   stubgen {...} -m mlir._mlir_libs._mlir.passmanager
# Local modifications:
#   * Relative imports for cross-module references.
#   * Add __all__

from typing import Any, Optional

from . import ir as _ir

__all__ = [
    "PassManager",
]

class PassManager:
    def __init__(self, context: Optional[_ir.Context] = None) -> None: ...
    def _CAPICreate(self) -> object: ...
    def _testing_release(self) -> None: ...
    def enable_ir_printing(self) -> None: ...
    def enable_verifier(self, enable: bool) -> None: ...
    @staticmethod
    def parse(pipeline: str, context: Optional[_ir.Context] = None) -> PassManager: ...
    def run(self, module: _ir.Module) -> None: ...
    @property
    def _CAPIPtr(self) -> object: ...



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""Experimental MLIR-PyTACO with sparse tensor support.

See http://tensor-compiler.org/ for TACO tensor compiler.

This module implements the Python classes for PyTACO index notation. These
include classes for data types, tensor dimension formats (aka mode formats),
tensor dimension orderings (aka mode ordering), tensor storage formats, and
tensors.

The PyTACO API doesn't follow the naming conversion required by the style guide
for this module. As such, we first implement the supporting classes and routines
following the style guide, and then define the type aliases and constants to
support the PyTACO API in the pytaco_api module.
"""

from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import abc
import ctypes
import dataclasses
import enum
import numpy as np
import functools
import operator
import os
import threading

# Import MLIR related modules.
from mlir import execution_engine
from mlir import ir
from mlir import runtime
from mlir.dialects import arith
from mlir.dialects import bufferization
from mlir.dialects import builtin
from mlir.dialects import func
from mlir.dialects import linalg
from mlir.dialects import sparse_tensor
from mlir.dialects.linalg.opdsl import lang

from . import mlir_pytaco_utils as utils

# TACO naming prefixes.
_TACO_INDEX_PREFIX = "i"
_TACO_TENSOR_PREFIX = "A"

# Bitwidths for pointers and indices.
_POINTER_BIT_WIDTH = 0
_INDEX_BIT_WIDTH = 0
# The entry point to the JIT compiled program.
_ENTRY_NAME = "main"

# Type aliases for type annotation.
_UnaryOp = Callable[[Any], Any]
_BinaryOp = Callable[[Any, Any], Any]
_ExprVisitor = Callable[..., None]
_ExprInfoDict = Dict["IndexExpr", "_ExprInfo"]
_LogicalOp = Callable[[bool, bool], bool]
_ModeFormatOp = Callable[["ModeFormat", "ModeFormat"], "ModeFormat"]
_SubtreeLeafChecker = Optional[Callable[..., bool]]


class Type(enum.Enum):
  """The data types supported by TACO.

  We use numpy data types to implement the enum data types.
  """
  INT8 = np.int8
  INT16 = np.int16
  INT32 = np.int32
  INT64 = np.int64
  FLOAT16 = np.float16
  FLOAT32 = np.float32
  FLOAT64 = np.float64
  COMPLEX64 = np.complex64
  COMPLEX128 = np.complex128


# All floating point type enums.
_FLOAT_TYPES = (Type.FLOAT16, Type.FLOAT32, Type.FLOAT64)
# All integral type enums.
_INT_TYPES = (Type.INT8, Type.INT16, Type.INT32, Type.INT64)
# All complex type enums.
_COMPLEX_TYPES = (Type.COMPLEX64, Type.COMPLEX128)
# Type alias for any numpy type used to implement the runtime support for the
# enum data types.
_AnyRuntimeType = Union[np.int8, np.int16, np.int32, np.int64, np.float16,
                        np.float32, np.float64, np.complex64, np.complex128]


@dataclasses.dataclass(frozen=True)
class DType:
  """The data type class.

  We support the TACO API dtype class with an alias of this class.

  The following methods are defined by the TACO API:
    is_float: Returns whether the data type represents a floating point value.
    is_int:   Returns whether the data type represents an integral value.

  Attributes:
    kind: A Type enum representing the data type.
    value: The numpy data type for the TACO data type.
  """
  kind: Type = Type.FLOAT32

  def is_float(self) -> bool:
    """Returns whether the data type represents a floating point value."""
    return self.kind in _FLOAT_TYPES

  def is_int(self) -> bool:
    """Returns whether the data type represents an integral value."""
    return self.kind in _INT_TYPES

  def is_complex(self) -> bool:
    """Returns whether the data type represents a complex value."""
    return self.kind in _COMPLEX_TYPES

  @property
  def value(self) -> _AnyRuntimeType:
    """Returns the numpy dtype for the data type."""
    return self.kind.value


def _dtype_to_mlir_str(dtype: DType) -> str:
  """Returns the MLIR string for the given dtype."""
  dtype_to_str = {
      Type.INT16: "i8",
      Type.INT16: "i16",
      Type.INT32: "i32",
      Type.INT64: "i64",
      Type.FLOAT16: "f16",
      Type.FLOAT32: "f32",
      Type.FLOAT64: "f64",
      Type.COMPLEX64: "complex<f32>",
      Type.COMPLEX128: "complex<f64>"
  }
  return dtype_to_str[dtype.kind]


def _nptype_to_taco_type(ty: np.dtype) -> DType:
  """Returns the TACO type for the given numpy type."""
  nptype_to_dtype = {
      np.int8: Type.INT8,
      np.int16: Type.INT16,
      np.int32: Type.INT32,
      np.int64: Type.INT64,
      np.float16: Type.FLOAT16,
      np.float32: Type.FLOAT32,
      np.float64: Type.FLOAT64,
      np.complex64: Type.COMPLEX64,
      np.complex128: Type.COMPLEX128
  }
  return DType(nptype_to_dtype[ty])


def _mlir_type_from_taco_type(dtype: DType) -> ir.Type:
  """Returns the MLIR type corresponding to the given TACO type."""
  dtype_to_irtype = {
      Type.INT8: ir.IntegerType.get_signless(8),
      Type.INT16: ir.IntegerType.get_signless(16),
      Type.INT32: ir.IntegerType.get_signless(32),
      Type.INT64: ir.IntegerType.get_signless(64),
      Type.FLOAT16: ir.F16Type.get(),
      Type.FLOAT32: ir.F32Type.get(),
      Type.FLOAT64: ir.F64Type.get(),
      Type.COMPLEX64: ir.ComplexType.get(ir.F32Type.get()),
      Type.COMPLEX128: ir.ComplexType.get(ir.F64Type.get())
  }
  return dtype_to_irtype[dtype.kind]

def _ctype_pointer_from_array(array: np.ndarray) -> ctypes.pointer:
  """Returns the ctype pointer for the given numpy array."""
  return ctypes.pointer(
      ctypes.pointer(runtime.get_ranked_memref_descriptor(array)))


class ModeFormat(enum.Enum):
  """The tensor dimension storage format class.

  We support the TACO API mode_format class with an alias of this class.

  In TACO, a tensor dimension is called a mode and the storage format for a
  tensor dimension is called a mode format.
  """
  DENSE = sparse_tensor.DimLevelType.dense
  COMPRESSED = sparse_tensor.DimLevelType.compressed


def _mode_format_operation(a: ModeFormat, b: ModeFormat,
                           op: _LogicalOp) -> ModeFormat:
  """Implements the given operator on ModeFormat."""
  return (ModeFormat.COMPRESSED
          if op(a == ModeFormat.COMPRESSED, b == ModeFormat.COMPRESSED) else
          ModeFormat.DENSE)


def _mode_format_estimator(op: _BinaryOp) -> _ModeFormatOp:
  """Produces a ModeFormat operator for the given binary operator.

  The ModeFormat operator is used as a heuristic to derive the destination
  dimension sparsity from the source dimension sparsity. In particular, if the
  binary operator produces a disjunction of the zero values from its source
  operands, such as the MUL operator, we return a ModeFormat operator that
  uses operator.or_. That is, we estimate that a dimension for the MUL
  operation result to be sparse if either of its source operands is sparse.

  On the other hand, if the binary operator produces a conjunction of the
  zero values from its source operands, such as the ADD operator, we return
  a ModeFormat operator that uses operator.and_. In this case, we estimate
  that a dimension for the ADD operation result to be sparse if both of its
  source operands are sparse.

  Args:
    op: A _BinaryOp object representing a supporting operator on tensors.

  Returns:
    A ModeFormatOp for estimating the destination dimension sparsity from
    the source dimension sparsity.
  """
  conjunction = functools.partial(_mode_format_operation, op=operator.and_)
  disjunction = functools.partial(_mode_format_operation, op=operator.or_)
  return conjunction if op(0, 1) != 0 else disjunction


def _all_instance_of(collection: Iterable, cls: Any) -> bool:
  """Returns true if all elements of the iterable is an instance of cls."""
  return all(isinstance(e, cls) for e in collection)


def _identity_ordering(rank: int) -> List[int]:
  """Returns the identity ordering for tensor of given rank."""
  return list(range(rank))


@dataclasses.dataclass(frozen=True)
class ModeOrdering:
  """The tensor dimension ordering class.

  We support the TACO API mode_ordering class with an alias of this class.

  Attributes:
    ordering: A list of integers representing the ordering of the tensor
      dimensions.
  """
  ordering: List[int]

  def __post_init__(self) -> None:
    """Verifies the value in ordering.

    Raises:
       ValueError: If ordering is not a list of integers.
    """
    if (not isinstance(self.ordering, list) or
        not _all_instance_of(self.ordering, int)):
      raise ValueError("Ordering must be a list of integers: "
                       f"{self.ordering}")
    # Check that ordering is a permutation of the dimension numbers.
    if sorted(self.ordering) != _identity_ordering(self.rank()):
      raise ValueError(f"Invalid ordering: {self.ordering} != "
                       f"permutation{_identity_ordering(self.rank())}.")

  def rank(self) -> int:
    """Returns the number of dimensions represented by the ordering."""
    return len(self.ordering)


@dataclasses.dataclass(frozen=True)
class ModeFormatPack:
  """The tensor dimension format class.

  We support the TACO API mode_format_pack class with an alias of this class.

  The storage format of a tensor contains one mode_format for each tensor
  dimension.

  Attributes:
    formats: A list of ModeFormat representing the storage format for each of
      the tensor dimension.
  """
  formats: List[ModeFormat]

  def __post_init__(self) -> None:
    """Verifies the value in formats.

    Raises:
       ValueError: If formats is not a list of ModeFormats.
    """
    if (not isinstance(self.formats, list) or
        not _all_instance_of(self.formats, ModeFormat)):
      raise ValueError("Formats must be a list of ModeFormat: "
                       f"{self.formats}")

  def rank(self) -> int:
    """Returns the number of dimensions represented by the format pack."""
    return len(self.formats)


@dataclasses.dataclass
class Format:
  """The tensor format class defined by the TACO API.

  Attributes:
    format_pack: A ModeFormatPack representing the storage format for the tensor
      dimensions.
    ordering: A ModeOrdering representing the tensor dimension ordering in the
      storage.
  """
  format_pack: ModeFormatPack
  ordering: Optional[ModeOrdering] = None

  def __post_init__(self) -> None:
    """Verifies and fixes up the values in format_pack and ordering.

    Verifies and fixes up the values in format_pack and ordering to supports the
    initializer syntax defined by the TACO API. If format_pack is a list of
    ModeFormat, replaces it with ModeFormatPack constructed from the list. If
    ordering is not provided, set ordering to the natural ordering for the rank
    corresponding to format_pack.

    Raises:
       ValueError: If format_pack is not an instance of ModeFormatPack or if
         ordering is not an instance of ModeOrdering.
    """
    if isinstance(self.format_pack, list):
      if not _all_instance_of(self.format_pack, ModeFormat):
        raise ValueError(f"Expected a list of ModeFormat: {self.format_pack}")
      self.format_pack = ModeFormatPack(self.format_pack)
    if not isinstance(self.format_pack, ModeFormatPack):
      raise ValueError(f"Expected ModeFormatpack: {self.format_pack}")

    if self.ordering is None:
      self.ordering = ModeOrdering(list(range(self.rank())))
    if isinstance(self.ordering, list):
      if not _all_instance_of(self.ordering, int):
        raise ValueError(f"Expected a list of integer: {self.ordering}")
      self.ordering = ModeOrdering(self.ordering)
    if not isinstance(self.ordering, ModeOrdering):
      raise ValueError(f"Expected ModeOrdering: {self.ordering}")

    if self.format_pack.rank() != self.ordering.rank():
      raise ValueError("Inconsistent ModeFormatPack and ModeOrdering: "
                       f"len({self.format_pack}) != "
                       f"len({self.ordering})")

  def rank(self) -> int:
    """Returns the number of dimensions represented by the format."""
    return self.format_pack.rank()

  def get_permutation_and_sparsity(self) -> Tuple[np.ndarray, np.ndarray]:
    """Constructs the numpy arrays for the permutation and sparsity."""
    perm = np.array(self.ordering.ordering, dtype=np.ulonglong)
    a = [0 if s == ModeFormat.DENSE else 1 for s in self.format_pack.formats]
    sparse = np.array(a, dtype=np.uint8)
    return (perm, sparse)

  def mlir_tensor_attr(self) -> Optional[sparse_tensor.EncodingAttr]:
    """Constructs the MLIR attributes for the tensor format."""
    order = (
        range(self.rank()) if
        (self.ordering is None) else self.ordering.ordering)
    mlir_storage_format = [f.value for f in self.format_pack.formats]
    return sparse_tensor.EncodingAttr.get(mlir_storage_format,
                                          ir.AffineMap.get_permutation(order),
                                          _POINTER_BIT_WIDTH, _INDEX_BIT_WIDTH)


def _make_format(formats: List[ModeFormat],
                 ordering: Optional[List[int]] = None) -> Format:
  """Constructs a format from a list of ModeFormat and an optional ordering.

  Args:
    formats: A list of ModeFormat, one for each dimension of a tensor.
    ordering: An optional list of integer, for the ordering of the tensor
      dimensions. When an ordering is not given, the identity ordering is used.

  Returns:
    A tensor format object.

  Raises:
    ValueError: If formats is not a list of ModeFormat or the length of formats
      is not consistent with the len of ordering.
  """
  ordering = ordering or _identity_ordering(len(formats))
  return Format(ModeFormatPack(formats), ModeOrdering(ordering))


class IndexExpr(abc.ABC):
  """The index notation base class.

  We support the TACO API index_expression class with an alias of this class.
  """

  def _verify_operand_and_build_expr(self, rhs, op: _BinaryOp) -> "_BinaryExpr":
    """Verifies the RHS operand and returns a binary expression.

    Args:
      rhs: The RHS of the binary operation, which could be any Python object
        from user inputs.
      op: A _BinaryOp object representing the binary operator.

    Raises:
      ValueError: If rhs is not an IndexExpr.
    """
    if not isinstance(rhs, IndexExpr):
      raise ValueError(f"Expected IndexExpr: {rhs}")
    return _BinaryExpr(op, self, rhs)

  def _build_unary_expr(self, op: _UnaryOp) -> "_UnaryExpr":
    """Build a unary expression.

    Args:
      op: A _UnaryOp object representing the unary operation.
    """
    return _UnaryExpr(op, self)

  def __add__(self, rhs) -> "_BinaryExpr":
    """Defines the operator +.

    Args:
      rhs: The value being added, which could be any Python object from user
        inputs.

    Returns:
      A _BinaryExpr object representing the operation.

    Raises:
      ValueError: If rhs is not an IndexExpr.
    """
    return self._verify_operand_and_build_expr(rhs, operator.add)

  def __mul__(self, rhs) -> "_BinaryExpr":
    """Defines the operator *.

    Args:
      rhs: The value being multiplied, which could be any Python object from
        user inputs.

    Returns:
      A _BinaryExpr object representing the operation.

    Raises:
      ValueError: If rhs is not an IndexExpr.
    """
    return self._verify_operand_and_build_expr(rhs, operator.mul)

  def __abs__(self) -> "_UnaryExpr":
    """Defines the operator abs.

    Returns:
      A _UnaryExpr object representing the operation.
    """
    return self._build_unary_expr(operator.abs)

  def __neg__(self) -> "_UnaryExpr":
    """Defines the operator neg.

    Returns:
      A _UnaryExpr object representing the operation.
    """
    return self._build_unary_expr(operator.neg)

  def __sub__(self, rhs) -> "_BinaryExpr":
    """Defines the operator -.

    Args:
      rhs: The value being subtracted, which could be any Python object from
        user inputs.

    Returns:
      A _BinaryExpr object representing the operation.

    Raises:
      ValueError: If rhs is not an IndexExpr.
    """
    return self._verify_operand_and_build_expr(rhs, operator.sub)

  @abc.abstractmethod
  def _visit(self,
             func: _ExprVisitor,
             args,
             *,
             leaf_checker: _SubtreeLeafChecker = None) -> None:
    """A post-order visitor.

    Args:
      func: A callable applied to each node in the expression tree.
      args: The variable-length arguments passed to the callable. These
        arguments are grouped as an iterable and will be unpacked before passing
        to the callable. This is to enable the keyword argument only syntax
        after this argument.
      leaf_checker: A callable object to identify nodes that should be treated
        as leaf nodes to support partial tree visiting.
    """
    pass

  @abc.abstractmethod
  def _emit_expression(
      self,
      expr_to_opnd: Dict["IndexExpr", lang.OperandDef],
      expr_to_info: _ExprInfoDict,
  ) -> lang.ScalarExpression:
    """Emits MLIR for the expression tree.

    Args:
      expr_to_opnd: A dictionary for looking up structured op input operands for
        the input nodes of the structured op.
      expr_to_info: A dictionary for looking up code generation information for
        expressions.

    Returns:
      A linalg dialect ScalarExpression for the expression.
    """
    pass

  @abc.abstractmethod
  def dtype(self) -> DType:
    """Returns the data type for the result of the expression."""
    pass

  def _emit_structured_op(self, expr_to_info: _ExprInfoDict) -> None:
    """Emits a structured op in the linalg dialect for the expression tree.

    We define a DefineOpcallable in the domain specific language for the linalg
    dialect and execute the callable to generate the structured op. Self is the
    root of the expression tree for the structured op.

    Args:
      expr_to_info: A dictionary for looking up code generation information for
        expressions.
    """
    op_info = expr_to_info[self].structop_info
    op_name = op_info.dst_name
    op_def = lang.LinalgOpDef(name=op_name)
    op_callable = lang.DefinedOpCallable(op_name, op_def)

    # Collect the input expression nodes for the structured op.
    expr_inputs = []
    self._visit(
        _gather_structured_op_input,
        (self, expr_to_info, expr_inputs),
        leaf_checker=_is_structured_op_leaf,
    )

    # Create a linalg structured op operand for each input expression node and
    # build a dictionary for looking up the information.
    expr_to_input_opnd = {
        e: _emit_structured_op_input(e, expr_to_info, op_def)
        for e in expr_inputs
    }

    # Emit the expression tree, which produces the value assigned to the
    # destination tensor.
    value = self._emit_expression(expr_to_input_opnd, expr_to_info)
    # Emit the structured op representation for the destination tensor.
    dst_opnd = _emit_operand(op_def, op_info.dst_indices, op_info.dst_name,
                             lang.OperandKind.OUTPUT_TENSOR)
    dst_dim_syms = _mlir_dimensions_from_index_vars(op_info.dst_indices)
    dst_use = lang.TensorUse(dst_opnd, dst_dim_syms)

    expr_info = expr_to_info[self]
    # If the structured op reduces some indices, explicitly represent the
    # reduction. This is done by generating a ReduceFn for the dimensions being
    # reduced in the linalg dialect and calling the function with the value
    # being reduced. We only support add reduction currently.
    if expr_info.reduce_indices:
      reduce_dims = _mlir_dimensions_from_index_vars(expr_info.reduce_indices)
      value = lang.ReduceFn.add[reduce_dims](value)

    # Emit the assignment as a comprehension in the linalg dialect.
    comp = lang.Comprehension((dst_use, value))
    op_def.comprehensions.append(comp)

    # The structured op in the linalg dialect requires an explicit
    # initialization for the destination tensor. Emit MLIR to initialize the
    # destination tensor.
    init = op_info.emit_tensor_init()

    # Collect MLIR values for the linalg input operands, with the assumption
    # that dictionary preserves the insertion order.
    args = [
        expr_to_info[expr].mlir_value
        for expr, opnd in expr_to_input_opnd.items()
    ]
    # Execute the DefineOpcallable object for the linalg dialect operation to
    # emit MLIR for the linalg structured op.
    expr_info.mlir_value = op_callable(*args, outs=[init])

  def _identify_structured_ops(
      self,
      expr_to_info: _ExprInfoDict,
      dst: "Tensor",
      dst_indices: Tuple["IndexVar", ...],
  ) -> List["IndexExpr"]:
    """Returns expression nodes for the roots of the identified structured ops.

    A structured op in the linalg dialect only supports reduction performed on
    the whole expression. If the expression tree contains reduction that are
    performed on part of the expression tree, the expression tree needs to be
    implemented with multiple structured ops. This routine identifies all the
    expression nodes that contain reduction as the root of structured ops in the
    linalg dialect.

    Args:
      expr_to_info: A dictionary for looking up code generation information for
        expressions.
      dst: A destination Tensor that accepts the value of the expression tree.
      dst_indices: The indices used by the destination index expression.

    Returns:
      An ordered list of IndexExpr for the root expressions of the structured
      ops, where child expressions go before parent expressions that use their
      results.
    """
    reduce_indices = tuple(
        set(expr_to_info[self].src_indices) - set(dst_indices))
    for reduce_index in reduce_indices:
      _mark_structured_op_root(self, reduce_index, expr_to_info)

    self._visit(_accumulate_reduce_indices, (expr_to_info,))
    structop_roots = []
    self._visit(_gather_structured_op, (expr_to_info, structop_roots))

    # Handle the root of the top level expression.
    if not structop_roots or structop_roots[-1] != self:
      # The top level expression is not a reduction. Add the top level
      # expression as a structured op root.
      structop_roots.append(self)

    # Use user specified information for the destination tensor to build an
    # _StructOpInfo for the top level expression.
    expr_to_info[self].structop_info = _StructOpInfo(dst_indices,
                                                     tuple(dst.shape),
                                                     dst.dtype, dst.name,
                                                     dst.format)

    return structop_roots

  def _validate_and_collect_expr_info(
      self,
      dst: "Tensor",
      dst_indices: Tuple["IndexVar", ...],
  ) -> _ExprInfoDict:
    """Propagates expression information for validation.

    Propagates the indices used by child expression nodes to parent expression
    nodes. Also collects and validates the sizes for the dimensions
    corresponding to the indices.

    Args:
      dst: A destination Tensor that accepts the value of the expression tree.
      dst_indices: The indices used by the destination index expression.

    Raises:
      ValueError if there is any inconsistency in indices or dimensional
      values.

    Returns:
      A dictionary of (IndexExpr, _ExprInfo).
    """
    expr_to_info = {}
    # Validate the expression tree and construct expression information.
    self._visit(_validate_and_collect_expr_info, (expr_to_info,))

    # Validate the destination dimension information.
    info = expr_to_info[self]
    index_to_dim_info = {i: d for i, d in zip(info.src_indices, info.dim_infos)}
    for i, d, in zip(dst_indices, dst.shape):
      if i not in index_to_dim_info:
        raise ValueError("Destination IndexVar not used in the "
                         f"source expression: {i}")
      else:
        if d != index_to_dim_info[i].dim and index_to_dim_info[i].dim != -1:
          raise ValueError(f"Inconsistent destination dimension for {i}: "
                           f"{d} vs {index_to_dim_info[i].dim}")

    return expr_to_info

  def _emit_assignment(
      self,
      module: ir.Module,
      dst: "Tensor",
      dst_indices: Tuple["IndexVar", ...],
      expr_to_info: _ExprInfoDict,
      input_accesses: List["Access"],
  ) -> None:
    """Emits an MLIR function for assigning the expression to a tensor."""
    input_types = [a.tensor.mlir_tensor_type() for a in input_accesses]

    # Build the kernel for the operations.
    with ir.InsertionPoint(module.body):

      @func.FuncOp.from_py_func(*input_types, name=_ENTRY_NAME)
      def linalg_funcop(*args):
        # Set up the mapping from the Access nodes to their MLIR values.
        for e, mlir in zip(input_accesses, args):
          expr_to_info[e].mlir_value = mlir

        # Emit structured ops in the linalg dialect to implement the assignment.
        for structop_root in self._identify_structured_ops(
            expr_to_info, dst, dst_indices):
          structop_root._emit_structured_op(expr_to_info)
          dst._record_stats(expr_to_info[structop_root].structop_info)

        # The function returns the MLIR value of the root expression.
        return expr_to_info[self].mlir_value

      linalg_funcop.func_op.attributes[
          "llvm.emit_c_interface"] = ir.UnitAttr.get()

  def get_input_accesses(self) -> List["Access"]:
    """Compute the list of input accesses for the expression."""
    input_accesses = []
    self._visit(_gather_input_accesses_index_vars, (input_accesses,))
    return input_accesses

  def compile(
      self,
      dst: "Tensor",
      dst_indices: Tuple["IndexVar", ...],
  ) -> execution_engine.ExecutionEngine:
    """Compiles the tensor assignment dst[dst_indices] = expression.

    Args:
      dst: The destination tensor.
      dst_indices: The tuple of IndexVar used to access the destination tensor.

    Returns:
      The execution engine for the tensor assignment.

    Raises:
      ValueError: If the expression is not proper or not supported.
    """
    expr_to_info = self._validate_and_collect_expr_info(dst, dst_indices)
    input_accesses = self.get_input_accesses()

    # Build and compile the module to produce the execution engine.
    with ir.Context(), ir.Location.unknown():
      module = ir.Module.create()
      self._emit_assignment(module, dst, dst_indices, expr_to_info,
                            input_accesses)
      engine = utils.compile_and_build_engine(module)

    return engine


class _AtomicCounter:
  """An atomic counter."""

  def __init__(self):
    self._counter = 0
    self._counter_lock = threading.Lock()

  def increment(self) -> int:
    """Increments the counter by one and returns the old value."""
    old_value = self._counter
    with self._counter_lock:
      self._counter = self._counter + 1
    return old_value


class IndexVar(IndexExpr):
  """The tensor index class.

  We support the TACO API index_var class with an alias of this class.

  An IndexVar object represents an index variable in tensor index notation.

  Attributes:
    name: A unique string name of the IndexVar.
  """
  _counter = _AtomicCounter()

  def __init__(self):
    id = self._counter.increment()
    self._name = f"{_TACO_INDEX_PREFIX}{id}"

  def __repr__(self) -> str:
    return f"IndexVar(name={repr(self._name)})"

  @property
  def name(self) -> str:
    """Returns the name of the IndexVar."""
    return self._name

  def _visit(self,
             func: _ExprVisitor,
             args,
             *,
             leaf_checker: _SubtreeLeafChecker = None) -> None:
    """A post-order visitor."""
    if leaf_checker:
      assert leaf_checker(self, *args)
    func(self, *args)

  def _emit_expression(
      self,
      expr_to_opnd: Dict[IndexExpr, lang.OperandDef],
      expr_to_info: _ExprInfoDict,
  ) -> lang.ScalarExpression:
    """Emits a index value casted to the data type of the tensor expression."""
    dim = getattr(lang.D, self.name)
    index = lang.index(dim)
    int_value = lang.TypeFn.cast_unsigned(lang.TV.I64, index)
    return lang.TypeFn.cast_unsigned(lang.T, int_value)

  def dtype(self) -> DType:
    """Returns the data type for the index value.

    This is unreachable for IndexVar.
    """
    assert 0


def get_index_vars(n: int) -> List[IndexVar]:
  """Returns a list of n IndexVar.

  This routine is defined by the TACO API.

  Args:
    n: An integer representing the number of IndexVar to get.

  Returns:
    A list of IndexVar.

  Raises:
    ValueError: if n is not a positive integer.
  """
  if not isinstance(n, int) or n <= 0:
    raise ValueError(f"Expected an integer: {n}.")
  # If lock contention ever becomes an issue, we could implement a bulk getter
  # that returns a range by only claiming the lock once.
  return [IndexVar() for i in range(n)]


def _mlir_symbols_from_index_vars(
    index_vars: Tuple[IndexVar, ...]) -> Tuple[lang.SymbolDef, ...]:
  """Returns a tuple of MLIR symbols for the given tuple of index_var."""
  return tuple(getattr(lang.S, i.name) for i in index_vars)


def _mlir_dimensions_from_index_vars(
    index_vars: Tuple[IndexVar, ...]) -> Tuple[lang.DimDef, ...]:
  """Returns a tuple of MLIR dimensions for the given tuple of index_var."""
  return tuple(getattr(lang.D, i.name) for i in index_vars)


def _mlir_tensor_type(
    dtype: DType, shape: Tuple[int, ...],
    attr: Optional[sparse_tensor.EncodingAttr]) -> ir.RankedTensorType:
  """Returns an MLIR tensor type.

  Args:
    dtype: An DType object for the element data type of the tensor.
    shape: A tuple of integer for the shape of the tensor.
    attr: An optional MLIR sparse tensor attribute, only provided if the tensor
      is a sparse tensor.

  Returns:
    An MLIR ranked tensor type.
  """
  ir_type = _mlir_type_from_taco_type(dtype)
  return ir.RankedTensorType.get(shape, ir_type, attr)


@dataclasses.dataclass(frozen=True)
class _StructOpInfo:
  """Information for generating a structured op in the linalg dialect.

  This information is associated with an expression node that serves as the
  root for an expression subtree implemented with a structured op.

  Attributes:
    dst_indices: A tuple of IndexVar, representing the result dimensions of the
      structured op. This is used to construct the temporary variable for the
      tensor to hold the structured op result.
    dst_dims: A tuple of int, representing the result shape of the structured
      op.
    dst_dtype: A DType representing the data type of the structured op result.
    dst_name: A string representing the name of the structured op result.
    dst_format: An optional Format object representing the destination tensor
      format. None represents a true dense tensor.
  """
  dst_indices: Tuple[IndexVar, ...]
  dst_dims: Tuple[int, ...]
  dst_dtype: DType
  dst_name: str
  dst_format: Optional[Format]

  def __post_init__(self) -> None:
    """Verifies the integrity of the attribute values."""
    assert len(self.dst_indices) == len(self.dst_dims)

  def emit_tensor_init(self) -> ir.RankedTensorType:
    """Returns an initialization for the destination tensor."""
    if self.dst_format is None or self.dst_format.rank() == 0:
      # Initialize the dense tensor.
      ir_type = _mlir_type_from_taco_type(self.dst_dtype)
      tensor = linalg.InitTensorOp(self.dst_dims, ir_type).result
      zero = arith.ConstantOp(ir_type, 0.0)
      return linalg.fill(zero, outs=[tensor])

    # Initialize the sparse tensor.
    mlir_type = _mlir_tensor_type(self.dst_dtype, self.dst_dims,
                                  self.dst_format.mlir_tensor_attr())
    index_type = ir.IndexType.get()
    return bufferization.AllocTensorOp(mlir_type, [], None, None)


class _Stats:
  """Information to describe how a tensor expression is implemented.

  Currently, we only record the temporary tensors introduced for splitting the
  original expression.
  """

  def __init__(self):
    self._temps = []

  def __repr__(self) -> str:
    return f"_Stats({repr(self._temps)})"

  def add_element(self, structop: _StructOpInfo):
    """Adds a temporary tensor."""
    self._temps.append(structop)

  def get_total(self) -> int:
    """Gets the total number of temporary tensors."""
    return len(self._temps)

  def _get_element(self, idx: int) -> _StructOpInfo:
    """Gets the ith temporary tensor."""
    assert idx < self.get_total()
    return self._temps[idx]

  def get_dimensions(self, idx: int) -> Tuple[int]:
    """Gets the dimensions for the ith temporary tensor."""
    return self._get_element(idx).dst_dims

  def get_formats(self, idx: int) -> Tuple[ModeFormat]:
    """Gets the ModeFormats for the ith temporary tensor."""
    return tuple(self._get_element(idx).dst_format.format_pack.formats)


class _SparseValueInfo(enum.Enum):
  """Describes how a sparse tensor value is stored.
  _UNPACKED: The sparse tensor value is stored as (coordnates, values) in
    Python.
  _PACKED: The sparse tensor value is stored as a C pointer to a packed MLIR
    sparse tensor.
  """
  _UNPACKED = 0
  _PACKED = 1


@dataclasses.dataclass(frozen=True)
class _Assignment:
  """Records an assignment to a tensor T as T[indices] = expression."""
  indices: Tuple["IndexVar", ...]
  expression: "IndexExpr"


class Tensor:
  """The tensor class.

  We support the TACO API tensor class with an alias of this class.

  This class is part of the TACO API with the following methods:
    insert: Inserts a value to the given coordinate in the tensor.
    to_array: Returns a numpy ndarray for the tensor.

  TACO API also defines the following arrtibutes for the class:
    dtype: A dtype object representing the data type of the tensor.
    format: A format object representing the storage format of the tensor.
    name: A string object representing the name of the tensor.
    order: An integral rank of the tensor.
    shape: A list of integers representing the shape of the tensor.

  We currently ignore the tensor dimension ordering for dense tensor.
  """
  _counter = _AtomicCounter()

  def _get_unique_name(self) -> str:
    """Returns a unique name for creating a new Tensor."""
    return f"{_TACO_TENSOR_PREFIX}{self._counter.increment()}"

  def _init_format(self, fmt: Union[ModeFormat, List[ModeFormat],
                                    Format]) -> None:
    """Process the fmt argument for the Tensor constructor.

    Args:
      fmt: This argument can be a ModeFormat, List[ModeFormat], or format. If
        this argument is a ModeFormat, uses this ModeFormat for all the tensor
        dimensions. If this argument is a list of ModeFormat, the len of the
        list should equal to the rank of the tensor. If this argument is a
        format, uses it for the format of the tensor.

    Raises:
      ValueError: If fmt is not one of the expected type or is inconsistent
        with the rank of the tensor. This is because fmt could be an users
        input.
    """
    if isinstance(fmt, ModeFormat):
      self._format = _make_format([fmt] * self.order)
    elif isinstance(fmt, list):
      if len(fmt) == self.order and isinstance(fmt[0], ModeFormat):
        self._format = _make_format(fmt)
      else:
        raise ValueError("Inconsistent shape and format: "
                         f"{self._shape}, {fmt}.")
    elif isinstance(fmt, Format):
      if fmt.rank() != self.order:
        raise ValueError("Inconsistent shape and format: "
                         f"{self._shape}, {fmt}.")
      else:
        self._format = fmt
    else:
      raise ValueError(f"Invalid format argument: {fmt}.")

  def __init__(self,
               value_or_shape: Optional[Union[List[int], Tuple[int, ...],
                                              complex, float, int]] = None,
               fmt: Optional[Union[ModeFormat, List[ModeFormat],
                                   Format]] = None,
               dtype: Optional[DType] = None,
               name: Optional[str] = None,
               is_dense: bool = False):
    """The tensor constructor interface defined by TACO API.

    Args:
      value_or_shape: This argument is optional and can be int, float,
        List[int], or Tuple[int, ...]. If this argument is an int or float,
        creates a scalar tensor and initializes it with the value. If this
        argument is a list or tuple of int, uses it as the shape to create a
        tensor.
      fmt: This argument can be a ModeFormat, List[ModeFormat], or format. If
        this argument is a ModeFormat, uses this ModeFormat for all the tensor
        dimensions. If this argument is a list of ModeFormat, the len of the
        list should equal to the rank of the tensor. If this argument is a
        format, uses it for the format of the tensor.
      dtype: An object of dtype, representing the data type of the tensor.
      name: A string name of the tensor. If a name is not given, creates a
        unique name for the tensor.
      is_dense: A boolean variable to indicate whether the tensor is a dense
        tensor without any sparsity annotation.

    Raises:
      ValueError: If there is any inconsistency among the input arguments.
    """
    # Take care of the argument default values common to both sparse tensors
    # and dense tensors.
    dtype = dtype or DType(Type.FLOAT32)
    self._name = name or self._get_unique_name()
    self._assignment = None
    self._engine = None
    self._sparse_value_location = _SparseValueInfo._UNPACKED
    self._dense_storage = None
    self._dtype = dtype

    if is_dense:
      assert (fmt is None)
      assert (isinstance(value_or_shape, tuple) or isinstance(
          value_or_shape, list)) and _all_instance_of(value_or_shape, int)
      self._shape = value_or_shape
      self._format = None
      return

    fmt = fmt or ModeFormat.COMPRESSED
    # We currently use _coords and _values to host the sparse tensor value with
    # COO format, and _dense_storage to host the dense tensor value. We don't
    # support the conversion between the two storages.
    self._coords = []
    self._values = []
    self._stats = _Stats()
    if value_or_shape is None or isinstance(value_or_shape, int) or isinstance(
        value_or_shape, float) or isinstance(value_or_shape, complex):
      # Create a scalar tensor and ignore the fmt parameter.
      self._shape = []
      self._format = _make_format([], [])
      if value_or_shape is not None:
        self._dense_storage = np.array(value_or_shape, dtype=self._dtype.value)
    elif (isinstance(value_or_shape, tuple) or isinstance(
        value_or_shape, list)) and _all_instance_of(value_or_shape, int):
      # Create a tensor with the specified shape and format.
      self._shape = list(value_or_shape)
      self._init_format(fmt)
    else:
      raise ValueError("Invalid first argument. "
                       "Must be a tuple or list for a shape or a single value"
                       f"if initializing a scalar tensor: {value_or_shape}.")

  def _set_packed_sparse_tensor(self, pointer: ctypes.c_void_p) -> None:
    """Records the MLIR sparse tensor pointer."""
    self._sparse_value_location = _SparseValueInfo._PACKED
    self._packed_sparse_value = pointer

  def is_unpacked(self) -> bool:
    """Returns true if the tensor value is not packed as MLIR sparse tensor."""
    return (self._sparse_value_location == _SparseValueInfo._UNPACKED)

  def unpack(self) -> None:
    """Unpacks the MLIR sparse tensor representation."""
    if self.is_dense() or self.is_unpacked():
      return

    # Use the output MLIR sparse tensor pointer to retrieve the COO-flavored
    # values and verify the values.
    rank, nse, shape, values, indices = utils.sparse_tensor_to_coo_tensor(
        self._packed_sparse_value, self._dtype.value)
    assert rank == self.order
    assert np.array_equal(self.shape, shape)
    assert nse == len(values)
    self._coords = indices
    self._values = values
    self._sparse_value_location = _SparseValueInfo._UNPACKED

  def __repr__(self) -> str:
    self._sync_value()
    self.unpack()
    value_str = (f"{repr(self._dense_storage)})" if self.is_dense() else
                 f"{repr(self._coords)} {repr(self._values)})")
    return (f"Tensor(_name={repr(self._name)} "
            f"_dtype={repr(self._dtype)} : ") + value_str

  def insert(self, coords: List[int], val: Union[complex, float, int]) -> None:
    """Inserts a value to the given coordinate.

    Args:
      coords: A list of integer coordinates. The length of the list must be the
        same as the rank of the tensor.
      val: A value being inserted. It is either an integral or a floating point
        value. This value will be converted to the data type of the tensor.

    Raises:
      ValueError: When there is any problem in the parameters.
    """
    if self.is_dense():
      raise ValueError("Insert method is not supported for dense tensors.")
    if self._assignment != None or not self.is_unpacked():
      raise ValueError(
          "Can't use Insert method for a tensor constructed from a file.")
    if not isinstance(coords, list):
      raise ValueError(f"Non list coordinate detected: {coords}.")
    if not _all_instance_of(coords, int):
      raise ValueError(f"Non integer coordinate detected: {coords}.")
    if (len(coords) != self.order or
        any([c < 0 or c >= self._shape[i] for i, c in enumerate(coords)])):
      raise ValueError("Invalid coordinate for rank: "
                       f"{self.order}, {coords}.")

    if not isinstance(val, int) and not isinstance(
        val, float) and not isinstance(val, complex):
      raise ValueError(f"Value is neither int nor float: {val}.")

    self._coords.append(tuple(coords))
    self._values.append(self._dtype.value(val))

  def is_dense(self) -> bool:
    """Returns true if the tensor doesn't have sparsity annotation."""
    return self.order == 0 or self._format is None

  def to_array(self) -> np.ndarray:
    """Returns the numpy array for the Tensor.

    This is currenly only implemented for dense Tensor.
    """
    if not self.is_dense():
      raise ValueError("Conversion from non-dense Tensor "
                       "to numpy array not supported yet.")

    self._sync_value()

    return self._dense_storage

  @staticmethod
  def from_array(array: np.ndarray) -> "Tensor":
    """Returns a dense tensor with the value copied from the input array.

    We currently only support the conversion of float32 and float64 numpy arrays
    to Tensor.

    Args:
      array: The numpy array that provides the data type, shape and value for
        the tensor.

    Returns:
      A Tensor object.

    Raises:
      ValueError if the data type of the numpy array is not supported.
    """
    if array.dtype != np.float32 and array.dtype != np.float64:
      raise ValueError(f"Expected floating point value type: {array.dtype}.")
    tensor = Tensor(
        array.shape,
        dtype=_nptype_to_taco_type(array.dtype.type),
        is_dense=True)
    tensor._dense_storage = np.copy(array)
    return tensor

  @staticmethod
  def from_coo(
      coordinates: List[Tuple[int, ...]],
      values: List[_AnyRuntimeType],
      fmt: Format,
      dtype: DType,
  ) -> "Tensor":
    """Converts coordinates and values to a sparse tensor representation.

    Args:
      coordinates: A list of coordinates with non-zero values.
      values: The non-zero values.
      fmt: The tensor storage format.
      dtype: The tensor element data type.

    Returns:
      A tensor with the given non-zero values and storage format. The shape of
      the tensor has the minimum size for each dimension to make the given
      coordinates valid.
    """
    assert (isinstance(coordinates, List) and
            _all_instance_of(coordinates, Tuple))
    assert (isinstance(values, List) and _all_instance_of(values, dtype.value))
    assert isinstance(fmt, Format)

    rank = fmt.rank()
    assert all(len(c) == rank and _all_instance_of(c, int) for c in coordinates)

    # Find the maximum coordinate value for each dimension.
    max_coordinate = list(map(max, zip(*coordinates)))
    # The size of each dimension is one more that such a maximum coordinate
    # value.
    shape = [c + 1 for c in max_coordinate]
    tensor = Tensor(shape, fmt, dtype=dtype)
    tensor._coords = coordinates
    tensor._values = values

    return tensor

  @staticmethod
  def from_file(
      filename: str,
      fmt: Format,
      dtype: DType,
  ) -> "Tensor":
    """Constructs a sparse tensor using the COO-flavored values from a file.

    Args:
      filename: A string for the name of the file that contains the sparse
        tensor data.
      fmt: The tensor storage format.
      dtype: The tensor element data type.

    Returns:
      A tensor with the given non-zero values and storage format. The tensor
      value is stored as an MLIR sparse tensor.
    """
    sparse_tensor, shape = utils.create_sparse_tensor(filename,
                                                      fmt.format_pack.formats,
                                                      _dtype_to_mlir_str(dtype))
    tensor = Tensor(shape.tolist(), fmt, dtype=dtype)
    tensor._set_packed_sparse_tensor(sparse_tensor)

    return tensor

  def to_file(self, filename: str) -> None:
    """Output the tensor value to a file.

    This method evaluates any pending assignment to the tensor and outputs the
    tensor value.

    Args:
      filename: A string file name.

    Raises:
       ValueError: If the tensor is dense, or an unpacked sparse tensor.
    """
    self._sync_value()

    if self.is_dense():
      raise ValueError("Writing dense tensors without sparsity annotation to "
                       "file is not supported.")

    if self.is_unpacked():
      raise ValueError("Writing unpacked sparse tensors to file is not "
                       "supported.")

    utils.output_sparse_tensor(self._packed_sparse_value, filename,
                               self._format.format_pack.formats,
                               _dtype_to_mlir_str(self._dtype))

  @property
  def dtype(self) -> DType:
    """Returns the data type for the Tensor."""
    return self._dtype

  @property
  def format(self) -> Format:
    """Returns the storage format for the Tensor."""
    return self._format

  @property
  def name(self) -> str:
    """Returns the name for the Tensor."""
    return self._name

  @property
  def order(self) -> int:
    """Returns the rank of the Tensor."""
    return len(self._shape)

  @property
  def shape(self) -> List[int]:
    """Returns the shape of the Tensor."""
    return self._shape

  def _verify_and_normalize_indices(self, indices) -> Tuple[IndexVar, ...]:
    """Verifies and normalizes the indices to access the tensor.

    Args:
      indices: The index expression used to access a tensor, which could be any
        Python object from user inputs.

    Returns:
      A tuple of IndexVar.

    Raises:
      ValueError: If indices is not 0 for scalar tensors, or not an IndexVar or
        a tuple of IndexVar for other tensors.
    """
    if self.order == 0:
      if not isinstance(indices, int) or indices != 0:
        raise ValueError(f"Expected 0 to index scalar tensors: {indices}")
      return ()

    if isinstance(indices, IndexVar):
      return (indices,)
    elif isinstance(indices, tuple) and _all_instance_of(indices, IndexVar):
      return indices

    raise ValueError(f"Expected IndexVars: {indices}")

  def __getitem__(self, key) -> "Access":
    """Verifies and processes a tensor access.

    In the tensor index notation, a tensor access T[i, j] is represented as
    retrieving a value with key (i, j) from the tensor object T in Python. This
    routine verifies the key for the tensor access and returns a tensor access
    object.

    Args:
      key: The key used to access the tensor, which could be any Python object
        from user inputs.

    Returns:
      The corresponding tensor access object.

    Raises:
      ValueError: If key is not an IndexVar or a tuple of IndexVar.
    """
    indices = self._verify_and_normalize_indices(key)
    return Access(self, indices)

  def __setitem__(self, key, value) -> None:
    """Verifies and processes a tensor assignment.

    In the tensor index notation, a tensor assignment "T[i, j] = ..." is
    represented as setting a value for a tensor object T via key (i, j) in
    Python. This routine verifies the key, evaluates the value, and assigns the
    value to the tensor.

    We only support assignment of dense tensor currently.

    Args:
      key: The key used to access the tensor, which could be any Python object
        from user inputs.
      value: The value assigned to the tensor, which could be any Python object
        from user inputs.

    Raises:
      ValueError: If tensor is not a dense tensor, or the key is not an IndexVar
        or a tuple of IndexVar, or the length of the indices is not the same as
        the rank of the tensor.
    """
    indices = self._verify_and_normalize_indices(key)
    if len(indices) != self.order:
      raise ValueError("Mismatch between indices and tensor rank: "
                       f"len({indices}) != {self.order}.")

    self._assignment = _Assignment(indices, value)
    self._engine = None

  def compile(self, force_recompile: bool = False) -> None:
    """Compiles the tensor assignment to an execution engine.

    Calling compile the second time does not do anything unless
    force_recompile is True.

    Args:
      force_recompile: A boolean value to enable recompilation, such as for the
        purpose of timing.

    Raises:
      ValueError: If the assignment is not proper or not supported.
    """
    if self._assignment is None or (self._engine is not None and
                                    not force_recompile):
      return

    self._engine = self._assignment.expression.compile(self,
                                                       self._assignment.indices)

  def compute(self) -> None:
    """Executes the engine for the tensor assignment.

    Raises:
      ValueError: If the assignment hasn't been compiled yet.
    """
    if self._assignment is None:
      return

    if self._engine is None:
      raise ValueError("Need to invoke compile() before invoking compute().")

    input_accesses = self._assignment.expression.get_input_accesses()
    # Gather the pointers for the input buffers.
    input_pointers = [a.tensor.ctype_pointer() for a in input_accesses]
    if self.is_dense():
      # The pointer to receive dense output is the first argument to the
      # execution engine.
      arg_pointers = [self.dense_dst_ctype_pointer()] + input_pointers
    else:
      # The pointer to receive the sparse tensor output is the last argument
      # to the execution engine and is a pointer to pointer of char.
      arg_pointers = input_pointers + [
          ctypes.pointer(ctypes.pointer(ctypes.c_char(0)))
      ]

    # Invoke the execution engine to run the module.
    self._engine.invoke(_ENTRY_NAME, *arg_pointers)

    # Retrieve the result.
    if self.is_dense():
      result = runtime.ranked_memref_to_numpy(arg_pointers[0][0])
      assert isinstance(result, np.ndarray)
      self._dense_storage = result
    else:
      self._set_packed_sparse_tensor(arg_pointers[-1][0])

    self._assignment = None
    self._engine = None

  def evaluate(self) -> None:
    """Evaluates the tensor assignment."""
    self.compile()
    self.compute()

  def _sync_value(self) -> None:
    """Updates the tensor value by evaluating the pending assignment."""
    if self._assignment is not None:
      self.evaluate()

  def mlir_tensor_type(self) -> ir.RankedTensorType:
    """Returns the MLIR type for the tensor."""
    mlir_attr = (None if (self._format is None or self.order == 0) else
                 self._format.mlir_tensor_attr())
    return _mlir_tensor_type(self._dtype, tuple(self._shape), mlir_attr)

  def dense_dst_ctype_pointer(self) -> ctypes.pointer:
    """Returns the ctypes pointer for the pointer to an MemRefDescriptor.

    For a dense tensor output, the MLIR compiler allocates the storage for
    the tensor. This routine returns the pointer to an MLIR MemRefDescriptor for
    receiving the tensor.
    """
    assert self.is_dense()
    mem_ref_desc = runtime.make_nd_memref_descriptor(
        self.order, np.ctypeslib.as_ctypes_type(self.dtype.value))()
    return ctypes.pointer(ctypes.pointer(mem_ref_desc))

  def ctype_pointer(self) -> ctypes.pointer:
    """Returns the ctypes pointer for the pointer to the input tensor."""
    if self.is_dense():
      if self._dense_storage is None:
        self._dense_storage = np.zeros(self._shape, self._dtype.value)
      return _ctype_pointer_from_array(self._dense_storage)

    if self.is_unpacked():
      shape = np.array(self._shape, np.int64)
      indices = np.array(self._coords, np.int64)
      values = np.array(self._values, self._dtype.value)
      perm, sparse = self.format.get_permutation_and_sparsity()
      ptr = utils.coo_tensor_to_sparse_tensor(shape, values, indices, perm,
                                              sparse)
    else:
      ptr = self._packed_sparse_value

    return ctypes.pointer(ctypes.cast(ptr, ctypes.c_void_p))

  def get_scalar_value(self) -> _AnyRuntimeType:
    """Returns the value for the scalar tensor.

    This method also evaluates the assignment to the tensor.

    Raises:
      ValueError: If the tensor is not a scalar.
    """
    if self.order != 0:
      raise ValueError(f"Expected a scalar tensor, got: rank={self.order}")

    self._sync_value()
    return self._dense_storage


  def get_coordinates_and_values(
      self) -> Tuple[List[Tuple[int, ...]], List[_AnyRuntimeType]]:
    """Returns the coordinates and values for the non-zero elements.

    This method also evaluates the assignment to the tensor and unpack the
    sparse tensor.
    """
    self._sync_value()

    if not self.is_dense():
      self.unpack()
      return (self._coords, self._values)

    if self.order == 0:
      return ([], self._dense_storage)

    # Coordinates for non-zero elements, grouped by dimensions.
    coords_by_dims = self._dense_storage.nonzero()
    # Coordinates for non-zero elements, grouped by elements.
    coords = np.transpose(coords_by_dims)
    values = self._dense_storage[coords_by_dims]
    return (coords, values)

  def _record_stats(self, structop: "_StructOpInfo"):
    """Collects information for temporary tensors."""
    # Exclude user specified destination tensors.
    if structop.dst_name == self.name:
      return

    self._stats.add_element(structop)


def _emit_operand(op_def: lang.LinalgOpDef, indices: Tuple[IndexVar, ...],
                  name: str, kind: lang.OperandKind) -> lang.OperandDef:
  """Emits an operand for a tensor access in the current linalg operation.

  Args:
    op_def: A LinalgOpDef representing the current linalg dialect operation.
    indices: A tuple of IndexVar used to access the tensor.
    name: A unique string name of the tensor.
    kind: An OperandKind for the operand.

  Returns:
    An OperandDef representing the operand.
  """
  dim_sym = _mlir_symbols_from_index_vars(indices)
  opnd = lang.OperandDef(kind, lang.T, dim_sym)
  op_def.add_operand(name, opnd)
  return opnd


@dataclasses.dataclass(frozen=True)
class _DimInfo:
  """Information for an operand dimension.

  Attributes:
    dim: An integer for the size of the dimension.
    mode_format: A ModeFormat for the dimension sparsity.
  """
  dim: int
  mode_format: ModeFormat


def _get_dummy_dim_info() -> _DimInfo:
  """Constructs the _DimInfo for an index used in tensor expressions."""
  return _DimInfo(-1, ModeFormat.DENSE)


@dataclasses.dataclass()
class _ExprInfo:
  """Expression information for validation and code generation.

  Attributes:
    src_indices: A tuple of IndexVar for the indices used by the tensors in the
      expression tree.
    dim_infos: A tuple of _DimInfo, representing the dimension information
      corresponding to the src_indices.
    reduce_indices: A set of IndexVar for the indices reduced by the expression.
    acc_reduce_indices: An accumulated set of IndexVar for the indices reduced
      by the expression and its children.
    structop_info: Information to support the code generation for a structured
      op in the linalg dialect, if the corresponding expression node is the root
      of a subtree for a structured op.
    mlir_value: The MLIR value generated for the structured op.
  """
  src_indices: Tuple[IndexVar, ...]
  dim_infos: Tuple[_DimInfo, ...]
  reduce_indices: Optional[Set[IndexVar]] = None
  acc_reduce_indices: Optional[Set[IndexVar]] = None
  structop_info: Optional[_StructOpInfo] = None
  mlir_value: Optional[ir.Value] = None

  def __post_init__(self) -> None:
    """Verifies and fix up attribute values.

    Verifies the consistency of the attributes and modifies the default values
    to support convenient initializer syntax.
    """
    assert len(self.src_indices) == len(self.dim_infos)
    self.reduce_indices = self.reduce_indices or set()
    self.acc_reduce_indices = self.acc_reduce_indices or set()


@dataclasses.dataclass(frozen=True)
class Access(IndexExpr):
  """The tensor access class.

  We support the TACO API access class with an alias of this class.

  Attributes:
    tensor: A Tensor being accessed.
    indices: A tuple of IndexVar, representing the indices used to access the
      Tensor.
  """
  tensor: Tensor
  indices: Tuple[IndexVar, ...]

  def __post_init__(self) -> None:
    """Verifies the tensor and indices for a tensor access.

    Raises:
       ValueError: If indices is not a list of IndexVar or the len of indices
       doesn't equal to the rank of the tensor.
    """
    if (not isinstance(self.indices, tuple) or
        not _all_instance_of(self.indices, IndexVar)):
      raise ValueError(f"Indices contain non IndexVar: {str(self.indices)}.")
    if self.tensor.order != len(self.indices):
      raise ValueError("Invalid indices for rank: "
                       f"str{self.tensor.order} != len({str(self.indices)}).")

  def __repr__(self) -> str:
    # The Tensor __repr__ method evaluates the pending assignment to the tensor.
    # We want to define the __repr__ method here to avoid such evaluation of the
    # tensor assignment.
    indices_str = ", ".join(map(lambda i: i.name, self.indices))
    return (f"Tensor({self.tensor.name}) " f"Indices({indices_str})")

  def _emit_expression(
      self,
      expr_to_opnd: Dict[IndexExpr, lang.OperandDef],
      expr_to_info: _ExprInfoDict,
  ) -> lang.ScalarExpression:
    """Emits a linalg dialect TensorUse expression for the tensor access."""
    assert self in expr_to_opnd
    dims = _mlir_dimensions_from_index_vars(self.indices)
    return lang.TensorUse(expr_to_opnd[self], dims)

  def _visit(self,
             func: _ExprVisitor,
             args,
             *,
             leaf_checker: _SubtreeLeafChecker = None) -> None:
    if leaf_checker:
      assert leaf_checker(self, *args)
    func(self, *args)

  def dtype(self) -> DType:
    return self.tensor.dtype


def _gather_input_accesses_index_vars(
    expr: IndexExpr,
    input_accesses: List[Access],
) -> None:
  """Collects Access nodes."""
  if isinstance(expr, Access) and expr not in input_accesses:
    input_accesses.append(expr)


def _op_ceil(__a: Any) -> Any:
  """A _UnaryOp object for operation ceil."""
  pass


def _op_floor(__a: Any) -> Any:
  """A _UnaryOp object for operation floor."""
  pass


def _op_unary_to_callable(op: _UnaryOp) -> lang.UnaryFnType:
  """Returns the linalg dialect function object for the given operation."""
  op_to_callable = {
      operator.abs: lang.UnaryFn.abs,
      operator.neg: lang.UnaryFn.negf,
      _op_ceil: lang.UnaryFn.ceil,
      _op_floor: lang.UnaryFn.floor,
  }
  return op_to_callable[op]


@dataclasses.dataclass(frozen=True)
class _UnaryExpr(IndexExpr):
  """The representation for a Unary operation.

  Attributes:
  op: A _UnaryOp representing the operation.
  a: An IndexExpr representing the operand for the operation.
  """
  op: _BinaryOp
  a: IndexExpr

  def __post_init__(self) -> None:
    """Verifies that the operand being added is an IndexExpr."""
    assert isinstance(self.a, IndexExpr)

  def _emit_expression(
      self,
      expr_to_opnd: Dict[IndexExpr, lang.OperandDef],
      expr_to_info: _ExprInfoDict,
  ) -> lang.ScalarExpression:
    """Emits the expression tree and returns the expression."""
    # The current expression node is an internal node of the structured op.
    if self not in expr_to_opnd:
      a = self.a._emit_expression(expr_to_opnd, expr_to_info)
      return _op_unary_to_callable(self.op)(a)

    # The current expression is a leaf node of the structured op. That is, it is
    # a temporary tensor generated by its child structured op.
    op_info = expr_to_info[self].structop_info
    assert op_info is not None
    dims = _mlir_dimensions_from_index_vars(op_info.dst_indices)
    return lang.TensorUse(expr_to_opnd[self], dims)

  def _visit(self,
             func: _ExprVisitor,
             args,
             *,
             leaf_checker: _SubtreeLeafChecker = None) -> None:
    """A post-order visitor."""
    if leaf_checker is None or not leaf_checker(self, *args):
      self.a._visit(func, args, leaf_checker=leaf_checker)
    func(self, *args)

  def dtype(self) -> DType:
    """Returns the data type of the operation."""
    return self.a.dtype()


def _op_to_callable(op: _BinaryOp) -> lang.BinaryFnType:
  """Returns the linalg dialect function object for the given operation."""
  op_to_callable = {
      operator.add: lang.BinaryFn.add,
      operator.sub: lang.BinaryFn.sub,
      operator.mul: lang.BinaryFn.mul,
  }
  return op_to_callable[op]

@dataclasses.dataclass(frozen=True)
class _BinaryExpr(IndexExpr):
  """The representation for a binary operation.

  Attributes:
  op: A _BinaryOp representing the binary operation.
  a: An IndexExpr representing the first operand of the operation.
  b: An IndexExpr representing the second operand of the operation.
  """
  op: _BinaryOp
  a: IndexExpr
  b: IndexExpr

  def __post_init__(self) -> None:
    """Verifies that the operands being added are IndexExpr."""
    assert isinstance(self.a, IndexExpr) and isinstance(self.b, IndexExpr)

  def _emit_expression(
      self,
      expr_to_opnd: Dict[IndexExpr, lang.OperandDef],
      expr_to_info: _ExprInfoDict,
  ) -> lang.ScalarExpression:
    """Emits the expression tree and returns the expression."""
    # The current expression node is an internal node of the structured op.
    if self not in expr_to_opnd:
      a = self.a._emit_expression(expr_to_opnd, expr_to_info)
      b = self.b._emit_expression(expr_to_opnd, expr_to_info)
      return _op_to_callable(self.op)(a, b)

    # The current expression is a leaf node of the structured op. That is, it is
    # a temporary tensor generated by its child structured op.
    op_info = expr_to_info[self].structop_info
    assert op_info is not None
    dims = _mlir_dimensions_from_index_vars(op_info.dst_indices)
    return lang.TensorUse(expr_to_opnd[self], dims)

  def _visit(self,
             func: _ExprVisitor,
             args,
             *,
             leaf_checker: _SubtreeLeafChecker = None) -> None:
    """A post-order visitor."""
    if leaf_checker is None or not leaf_checker(self, *args):
      self.a._visit(func, args, leaf_checker=leaf_checker)
      self.b._visit(func, args, leaf_checker=leaf_checker)
    func(self, *args)

  def dtype(self) -> DType:
    """Returns the data type of the binary operation."""
    return self.a.dtype()


def _validate_and_collect_dim_info(
    index_to_dim_info: Dict[IndexVar, _DimInfo],
    indices: Tuple[IndexVar, ...],
    dim_infos: Tuple[_DimInfo, ...],
    expr: _BinaryExpr,
) -> None:
  """Validates and collects the dimension information for an index notation.

  Validates (indices, dim_infos) against the information collected from other
  source operands and is represented by index_to_dim_info. In particular, we
  ensure that each IndexVar corresponds to only one dimension size. We also
  aggregate the new information represented in (indices, dim_infos) to
  index_to_dim_info.

  Args:
    index_to_dim: A dictionary of (IndexVar, _DimInfo) collected from the
      previous operands.
    indices: The IndexVars to be validated.
    dim_infos: The dimension information for the IndexVars to be validated.
    expr: The binary expression where (indices, dim_infos) is used.

  Raises:
    ValueError if there is any problem in the IndexVars or dimensional values.
  """
  assert len(indices) == len(dim_infos)
  for i, d in zip(indices, dim_infos):
    if i not in index_to_dim_info:
      index_to_dim_info[i] = d
    else:
      dim = index_to_dim_info[i].dim
      if dim == -1 or d.dim == -1:
        dim = dim if dim != -1 else d.dim
      elif dim != d.dim:
        raise ValueError(f"Inconsistent source dimension for {i}: "
                         f"{d.dim} vs {dim}")
      mode_format = _mode_format_estimator(expr.op)(
          index_to_dim_info[i].mode_format, d.mode_format)
      index_to_dim_info[i] = _DimInfo(d.dim, mode_format)


def _validate_and_collect_expr_info(
    expr: IndexExpr,
    expr_to_info: _ExprInfoDict,
) -> None:
  """Validates dimension information and constructs _ExprInfo.

  Validates that dimensional values for the same IndexVar are the same. Collects
  a list of IndexVar used by the expression and their corresponding dimensional
  values. Constructs an _ExprInfo object to record the information for the
  IndexExpr.

  This routine is passed to the post-order visitor as an _ExprVisitor object.

  Args:
    expr: The IndexExpr being validated.
    expr_to_info: The dictionary of (IndexExpr, _ExprInfo) for recording the
      expression information.

  Raises:
    ValueError if there is any problem in the IndexVars or dimensional values.
  """
  # Objects of class Access can be shared by different expressions. Avoid
  # processing Access objects multiple times by skipping the processing if expr
  # is already in the dictionary.
  if expr in expr_to_info:
    return

  if isinstance(expr, IndexVar):
    src_indices = expr,  # A tuple with one element.
    dim_infos = _get_dummy_dim_info(),  # A tuple with one element.
  elif isinstance(expr, Access):
    src_indices = expr.indices
    src_dims = tuple(expr.tensor.shape)
    if expr.tensor.format is None:
      # Treat each dimension of a dense tensor as DENSE for the purpose of
      # calculating temporary tensor storage format.
      mode_formats = tuple([ModeFormat.DENSE] * len(src_dims))
    else:
      mode_formats = tuple(expr.tensor.format.format_pack.formats)
    assert len(src_dims) == len(mode_formats)
    dim_infos = tuple([_DimInfo(d, m) for d, m in zip(src_dims, mode_formats)])
  elif isinstance(expr, _UnaryExpr):
    a_info = expr_to_info[expr.a]
    index_to_dim_info = {
        i: d for i, d in zip(a_info.src_indices, a_info.dim_infos)
    }
    # Here we rely on the fact that dictionaries keep the insertion order for
    # keys and values.
    src_indices = tuple(index_to_dim_info.keys())
    dim_infos = tuple(index_to_dim_info.values())
  else:
    assert isinstance(expr, _BinaryExpr)
    a_info = expr_to_info[expr.a]
    index_to_dim_info = {
        i: d for i, d in zip(a_info.src_indices, a_info.dim_infos)
    }
    b_info = expr_to_info[expr.b]
    _validate_and_collect_dim_info(index_to_dim_info, b_info.src_indices,
                                   b_info.dim_infos, expr)
    # Here we rely on the fact that dictionaries keep the insertion order for
    # keys and values.
    src_indices = tuple(index_to_dim_info.keys())
    dim_infos = tuple(index_to_dim_info.values())

  expr_to_info[expr] = _ExprInfo(src_indices, dim_infos)


def _mark_structured_op_root(
    expr: IndexExpr,
    reduce_index: IndexVar,
    expr_to_info: _ExprInfoDict,
) -> None:
  """Identifies the root expression for a structured op in the linalg dialect.

  An linalg structured op can only perform reduction on the whole expression.
  For a TACO tensor algebra expression, the reduction on an IndexVar is done at
  the smallest expression that contains all the uses of the IndexVar. If such an
  expression is only part of the whole expression, we need to split this
  sub-expression tree out from its parent and implement the sub-expression as a
  structured op.

  This routine identifies the root expression node for performing a reduction on
  the given IndexVar. If the reduction of the given IndexVar should be performed
  on expression X, then the IndexVar is added to expr_to_info[X].reduce_indices

  Args:
    expr: The root IndexExpr for the tensor algebra expression.
    reduce_index: The IndexVar which we want to find out the proper expression
      to perform a reduction.
    expr_to_info: The dictionary to look up _ExprInfo for IndexExpr.

  Raises:
      ValueError: If the expression is not proper or not supported.
  """
  expr_info = expr_to_info[expr]
  if isinstance(expr, Access):
    # Handle simple reduction expression in the format of A[i] = B[i, j].
    if reduce_index in expr_info.src_indices:
      expr_info.reduce_indices.add(reduce_index)
    return
  elif isinstance(expr, IndexVar):
    # A[i] = B[i] + j is not allowed.
    raise ValueError(f"IndexVar is not part of the iteration domain: {expr}.")

  assert (isinstance(expr, _BinaryExpr))
  a_info = expr_to_info[expr.a]
  b_info = expr_to_info[expr.b]

  if reduce_index in a_info.src_indices and reduce_index in b_info.src_indices:
    expr_info.reduce_indices.add(reduce_index)
    return

  if reduce_index in a_info.src_indices:
    _mark_structured_op_root(expr.a, reduce_index, expr_to_info)
  elif reduce_index in b_info.src_indices:
    _mark_structured_op_root(expr.b, reduce_index, expr_to_info)
  else:
    assert False, "Unreachable path"


def _accumulate_reduce_indices(
    expr: IndexExpr,
    expr_to_info: _ExprInfoDict,
) -> None:
  """Propagates reduction indices from child expressions to parent expressions.

  This routine is passed to the post-order visitor as an _ExprVisitor object.

  Args:
    expr: The IndexExpr being visited.
    expr_to_info: The dictionary of (IndexExpr, _ExprInfo) for recording the
      expression information.
  """
  assert expr in expr_to_info
  expr_info = expr_to_info[expr]

  if isinstance(expr, _BinaryExpr):
    a_info = expr_to_info[expr.a]
    b_info = expr_to_info[expr.b]
    expr_info.acc_reduce_indices = (
        a_info.acc_reduce_indices | b_info.acc_reduce_indices
        | expr_info.reduce_indices)
  elif isinstance(expr, _UnaryExpr):
    a_info = expr_to_info[expr.a]
    expr_info.acc_reduce_indices = (
        a_info.acc_reduce_indices | expr_info.reduce_indices)
  elif isinstance(expr, IndexVar):
    # If an IndexVar is reducing itself, it means the IndexVar is outside the
    # iteration domain. This usage is now allowed and we should emit an error
    # before reaching here.
    assert not expr_info.reduce_indices
  else:
    assert isinstance(expr, Access)
    # Handle simple reduction expression in the format of A[i] = B[i, j].
    expr_info.acc_reduce_indices = expr_info.reduce_indices



def _gather_structured_op(
    expr: IndexExpr,
    expr_to_info: _ExprInfoDict,
    structop_roots: List[IndexExpr],
) -> None:
  """Adds structured op root expression information to structop_roots.

  This routine is passed to the post-order visitor as an _ExprVisitor object.

  Args:
    expr: The IndexExpr being visited.
    expr_to_info: The dictionary to look up _ExprInfo for IndexExpr.
    structop_roots: The resulting list of IndexExpr that are the roots for
      linalg structured ops.
  """
  if not expr_to_info[expr].reduce_indices:
    return

  # If the expression is the root for reducing some indices, collect the indices
  # and dimensions for the reduction result.
  dst_indices = []
  dst_dims = []
  mode_fmts = []
  for i, d in zip(expr_to_info[expr].src_indices, expr_to_info[expr].dim_infos):
    if i not in expr_to_info[expr].acc_reduce_indices:
      dst_indices.append(i)
      dst_dims.append(d.dim)
      mode_fmts.append(d.mode_format)

  # Add the information to the dictionary.
  op_info = _StructOpInfo(
      tuple(dst_indices),
      tuple(dst_dims),
      expr.dtype(),
      f"temp{len(structop_roots)}",
      _make_format(mode_fmts),
  )
  expr_to_info[expr].structop_info = op_info

  # Add the expression to the list of structured op roots.
  structop_roots.append(expr)


def _is_structured_op_leaf(
    expr: IndexExpr,
    root: IndexExpr,
    expr_to_info: _ExprInfoDict,
    *unused_args,
) -> bool:
  """Returns true iff the expression is a leaf node for a structured op.

  The root of a structured op is a leaf of its parent structured op that uses
  its result. An expression node is a leaf node for the current structured op if
  it is an Access node or the root for a structured op that is not the current
  structured op.

  This routine is passed to the post-order visitor as a _SubtreeLeafChecker
  object. Because the post-order visitor pass the same parameters to both
  _SubtreeLeafChecker and _ExprVisitor, this routine may received unused
  parameters.

  Args:
    expr: The IndexExpr being visited.
    root: The root of the current structured op.
    expr_to_info: The dictionary to look up _ExprInfo for IndexExpr.

  Returns:
    True if the current IndexExpr is a leaf for the current structured op.
  """
  return (expr != root and
          expr_to_info[expr].structop_info is not None) or isinstance(
              expr, Access) or isinstance(expr, IndexVar)


def _gather_structured_op_input(
    expr: IndexExpr,
    root: IndexExpr,
    expr_to_info: _ExprInfoDict,
    structop_inputs: List[IndexExpr],
) -> None:
  """Adds the IndexExpr to structop_inputs if it is an input.

  If the current IndexExpr is an input for the current structured op, adds it to
  structop_inputs. The current IndexExpr is an input if it is an Access node or
  if it is the root for a structured op that is not the current structured op.

  This routine is passed to the post-order visitor as an _ExprVisitor object.

  Args:
    expr: The IndexExpr being visited.
    root: The root of the current structured op.
    expr_to_info: The dictionary to look up _ExprInfo for IndexExpr.
    structop_inputs: The resulting list of IndexExpr that provide input to the
      current structured op.
  """
  if ((expr != root or isinstance(expr, Access)) and
      expr not in structop_inputs) and (isinstance(expr, Access) or
                                        (expr in expr_to_info and
                                         expr_to_info[expr].structop_info)):
    structop_inputs.append(expr)


def _emit_structured_op_input(
    expr: IndexExpr,
    expr_to_info: _ExprInfoDict,
    op_def: lang.LinalgOpDef,
) -> lang.OperandDef:
  """Emits OperandDef in the linalg dialect for the input IndexExpr.

  Args:
    expr: The input IndexExpr for the current structured op.
    expr_to_info: The dictionary to look up _ExprInfo for IndexExpr.
    op_def: The linalg operation for the current structured op.

  Returns:
    An OperandDef in the linalg dialect for the input IndexExpr.
  """
  op_info = expr_to_info[expr].structop_info
  if op_info and not isinstance(expr, Access):
    # The input is a temporary tensor produced by another structured op.
    indices = op_info.dst_indices
    name = op_info.dst_name
  else:
    # The input is a user provided tensor.
    assert isinstance(expr, Access)
    indices = expr.indices
    name = expr.tensor.name

  dim_sym = _mlir_symbols_from_index_vars(indices)
  opnd = lang.OperandDef(lang.OperandKind.INPUT_TENSOR, lang.T, dim_sym)
  op_def.add_operand(name, opnd)
  return opnd


def _check_and_build_unary(a: Access, op: _UnaryOp) -> "_UnaryExpr":
  """Build a unary operation ceil.

    Args:
      a: The operand, which could be any Python object from user inputs.
      op: An _UnaryOp object representing the operation.

    Returns:
      A _UnaryExpr object representing the operation.

    Raises:
      ValueError: If a is not an IndexExpr.
    """
  if not isinstance(a, Access):
    raise ValueError(f"Expected an Access Operand: {a}")
  return a._build_unary_expr(op)


def ceil(a: Access) -> "_UnaryExpr":
  """Defines the operation ceil.

    Args:
      a: The operand, which could be any Python object from user inputs.

    Returns:
      A _UnaryExpr object representing the operation.

    Raises:
      ValueError: If a is not an IndexExpr.
    """
  return _check_and_build_unary(a, _op_ceil)


def floor(a: Access) -> "_UnaryExpr":
  """Defines the operation floor.

    Args:
      a: The operand, which could be any Python object from user inputs.

    Returns:
      A _UnaryExpr object representing the operation.

    Raises:
      ValueError: If a is not an IndexExpr.
    """
  return _check_and_build_unary(a, _op_floor)



"""
Test that target var can resolve complex DWARF expressions.
"""

import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class targetCommandTestCase(TestBase):

    @skipUnlessDarwin
    @skipIfDarwinEmbedded           # needs x86_64
    @skipIf(debug_info="gmodules")  # not relevant
    @skipIf(compiler="clang", compiler_version=['<', '7.0'])
    def testTargetVarExpr(self):
        self.build()
        lldbutil.run_to_name_breakpoint(self, 'main')
        self.expect("target variable i", substrs=['i', '42'])
        self.expect("target variable var", patterns=['\(incomplete \*\) var = 0[xX](0)*dead'])
        self.expect("target variable var[0]", error=True, substrs=["can't find global variable 'var[0]'"])



"""
Test lldb data formatter subsystem.
"""

from __future__ import print_function


import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbbench import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class TestBenchmarkContinue(BenchBase):

    @benchmarks_test
    def test_run_command(self):
        """Benchmark different ways to continue a process"""
        self.build()
        self.data_formatter_commands()

    def setUp(self):
        # Call super's setUp().
        BenchBase.setUp(self)

    def data_formatter_commands(self):
        """Benchmark different ways to continue a process"""
        self.runCmd("file "+self.getBuildArtifact("a.out"),
                    CURRENT_EXECUTABLE_SET)

        bkpt = self.target().FindBreakpointByID(
            lldbutil.run_break_set_by_source_regexp(
                self, "// break here"))

        self.runCmd("run", RUN_SUCCEEDED)

        # The stop reason of the thread should be breakpoint.
        self.expect("thread list", STOPPED_DUE_TO_BREAKPOINT,
                    substrs=['stopped',
                             'stop reason = breakpoint'])

        # This is the function to remove the custom formats in order to have a
        # clean slate for the next test case.
        def cleanup():
            self.runCmd('type format clear', check=False)
            self.runCmd('type summary clear', check=False)
            self.runCmd('type filter clear', check=False)
            self.runCmd('type synth clear', check=False)
            self.runCmd(
                "settings set target.max-children-count 256",
                check=False)

        # Execute the cleanup function during test case tear down.
        self.addTearDownHook(cleanup)

        runCmd_sw = Stopwatch()
        lldbutil_sw = Stopwatch()

        for i in range(0, 15):
            runCmd_sw.start()
            self.runCmd("continue")
            runCmd_sw.stop()

        for i in range(0, 15):
            lldbutil_sw.start()
            lldbutil.continue_to_breakpoint(self.process(), bkpt)
            lldbutil_sw.stop()

        print("runCmd: %s\nlldbutil: %s" % (runCmd_sw, lldbutil_sw))



import lldb
import lldb.formatters
import lldb.formatters.synth


class SyntheticChildrenProvider(
        lldb.formatters.synth.PythonObjectSyntheticChildProvider):

    def __init__(self, value, internal_dict):
        lldb.formatters.synth.PythonObjectSyntheticChildProvider.__init__(
            self, value, internal_dict)

    def make_children(self):
        return [("ID", 123456),
                ("Name", "Enrico"),
                ("Rate", 1.25)]



"""Test custom import command to import files by path."""



import lldb
from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
from lldbsuite.test import lldbutil


class ImportTestCase(TestBase):

    @add_test_categories(['pyapi'])
    @no_debug_info_test
    def test_import_command(self):
        """Import some Python scripts by path and test them"""
        self.run_test()

    def run_test(self):
        """Import some Python scripts by path and test them."""

        # This is the function to remove the custom commands in order to have a
        # clean slate for the next test case.
        def cleanup():
            self.runCmd('command script delete foo2cmd', check=False)
            self.runCmd('command script delete foocmd', check=False)
            self.runCmd('command script delete foobarcmd', check=False)
            self.runCmd('command script delete barcmd', check=False)
            self.runCmd('command script delete barothercmd', check=False)
            self.runCmd('command script delete TPcommandA', check=False)
            self.runCmd('command script delete TPcommandB', check=False)

        # Execute the cleanup function during test case tear down.
        self.addTearDownHook(cleanup)

        self.runCmd("command script import ./foo/foo.py --allow-reload")
        self.runCmd("command script import ./foo/foo2.py --allow-reload")
        self.runCmd("command script import ./foo/bar/foobar.py --allow-reload")
        self.runCmd("command script import ./bar/bar.py --allow-reload")

        self.expect("command script import ''",
                    error=True, startstr="error: module importing failed: empty path")
        self.expect("command script import ./nosuchfile.py",
                    error=True, startstr="error: module importing failed: invalid pathname './nosuchfile.py'")
        self.expect("command script import ./nosuchfolder/",
                    error=True, startstr="error: module importing failed: invalid pathname './nosuchfolder/'")
        self.expect("command script import ./foo/foo.py", error=False)
        self.runCmd("command script import --allow-reload ./thepackage")
        self.expect("TPcommandA", substrs=["hello world A"])
        self.expect("TPcommandB", substrs=["hello world B"])

        self.runCmd("script import dummymodule")
        self.expect("command script import ./dummymodule.py", error=False)
        self.expect(
            "command script import --allow-reload ./dummymodule.py",
            error=False)

        self.runCmd("command script add -f foo.foo_function foocmd")
        self.runCmd("command script add -f foobar.foo_function foobarcmd")
        self.runCmd("command script add -f bar.bar_function barcmd")
        self.expect("foocmd hello",
                    substrs=['foo says', 'hello'])
        self.expect("foo2cmd hello",
                    substrs=['foo2 says', 'hello'])
        self.expect("barcmd hello",
                    substrs=['barutil says', 'bar told me', 'hello'])
        self.expect("barothercmd hello",
                    substrs=['barutil says', 'bar told me', 'hello'])
        self.expect("foobarcmd hello",
                    substrs=['foobar says', 'hello'])



"""
Test stop hooks
"""


from lldbsuite.test.decorators import *
from lldbsuite.test.lldbtest import *
import lldbvscode_testcase


class TestVSCode_stop_hooks(lldbvscode_testcase.VSCodeTestCaseBase):

    @skipIfRemote
    def test_stop_hooks_before_run(self):
        '''
            Test that there is no race condition between lldb-vscode and
            stop hooks executor
        '''
        program = self.getBuildArtifact("a.out")
        preRunCommands = ['target stop-hook add -o help']
        self.build_and_launch(program, stopOnEntry=True, preRunCommands=preRunCommands)

        # The first stop is on entry.
        self.continue_to_next_stop()

        breakpoint_ids = self.set_function_breakpoints(['main'])
        # This request hangs if the race happens, because, in that case, the
        # command interpreter is in synchronous mode while lldb-vscode expects
        # it to be in asynchronous mode, so, the process doesn't send the stop
        # event to "lldb.Debugger" listener (which is monitored by lldb-vscode).
        self.continue_to_breakpoints(breakpoint_ids)

        self.continue_to_exit()



#!/usr/bin/env python3

"""Generate many skeletal functions with a thick call graph spanning a
large address space to induce lld to create branch-islands for arm64.

"""
from __future__ import print_function
import random
import argparse
import string
from pprint import pprint
from math import factorial
from itertools import permutations

# This list comes from libSystem.tbd and contains a sizeable subset
# of dylib calls available for all MacOS target archs.
libSystem_calls = (
  "__CurrentRuneLocale", "__DefaultRuneLocale", "__Exit", "__NSGetArgc",
  "__NSGetArgv", "__NSGetEnviron", "__NSGetMachExecuteHeader",
  "__NSGetProgname", "__PathLocale", "__Read_RuneMagi", "___Balloc_D2A",
  "___Bfree_D2A", "___ULtod_D2A", "____mb_cur_max", "____mb_cur_max_l",
  "____runetype", "____runetype_l", "____tolower", "____tolower_l",
  "____toupper", "____toupper_l", "___add_ovflpage", "___addel",
  "___any_on_D2A", "___assert_rtn", "___b2d_D2A", "___big_delete",
  "___big_insert", "___big_keydata", "___big_return", "___big_split",
  "___bigtens_D2A", "___bt_close", "___bt_cmp", "___bt_defcmp",
  "___bt_defpfx", "___bt_delete", "___bt_dleaf", "___bt_fd",
  "___bt_free", "___bt_get", "___bt_new", "___bt_open", "___bt_pgin",
  "___bt_pgout", "___bt_put", "___bt_ret", "___bt_search", "___bt_seq",
  "___bt_setcur", "___bt_split", "___bt_sync", "___buf_free",
  "___call_hash", "___cleanup", "___cmp_D2A", "___collate_equiv_match",
  "___collate_load_error", "___collate_lookup", "___collate_lookup_l",
  "___copybits_D2A", "___cxa_atexit", "___cxa_finalize",
  "___cxa_finalize_ranges", "___cxa_thread_atexit", "___d2b_D2A",
  "___dbpanic", "___decrement_D2A", "___default_hash", "___default_utx",
  "___delpair", "___diff_D2A", "___dtoa", "___expand_table",
  "___fflush", "___fgetwc", "___find_bigpair", "___find_last_page",
  "___fix_locale_grouping_str", "___fread", "___free_ovflpage",
  "___freedtoa", "___gdtoa", "___gdtoa_locks", "___get_buf",
  "___get_page", "___gethex_D2A", "___getonlyClocaleconv",
  "___hash_open", "___hdtoa", "___hexdig_D2A", "___hexdig_init_D2A",
  "___hexnan_D2A", "___hi0bits_D2A", "___hldtoa", "___i2b_D2A",
  "___ibitmap", "___increment_D2A", "___isctype", "___istype",
  "___istype_l", "___ldtoa", "___libc_init", "___lo0bits_D2A",
  "___log2", "___lshift_D2A", "___maskrune", "___maskrune_l",
  "___match_D2A", "___mb_cur_max", "___mb_sb_limit", "___memccpy_chk",
  "___memcpy_chk", "___memmove_chk", "___memset_chk", "___mult_D2A",
  "___multadd_D2A", "___nrv_alloc_D2A", "___opendir2", "___ovfl_delete",
  "___ovfl_get", "___ovfl_put", "___pow5mult_D2A", "___put_page",
  "___quorem_D2A", "___ratio_D2A", "___rec_close", "___rec_delete",
  "___rec_dleaf", "___rec_fd", "___rec_fmap", "___rec_fpipe",
  "___rec_get", "___rec_iput", "___rec_open", "___rec_put",
  "___rec_ret", "___rec_search", "___rec_seq", "___rec_sync",
  "___rec_vmap", "___rec_vpipe", "___reclaim_buf", "___rshift_D2A",
  "___rv_alloc_D2A", "___s2b_D2A", "___sF", "___sclose", "___sdidinit",
  "___set_ones_D2A", "___setonlyClocaleconv", "___sflags", "___sflush",
  "___sfp", "___sfvwrite", "___sglue", "___sinit", "___slbexpand",
  "___smakebuf", "___snprintf_chk", "___snprintf_object_size_chk",
  "___split_page", "___sprintf_chk", "___sprintf_object_size_chk",
  "___sread", "___srefill", "___srget", "___sseek", "___stack_chk_fail",
  "___stack_chk_guard", "___stderrp", "___stdinp", "___stdoutp",
  "___stpcpy_chk", "___stpncpy_chk", "___strcat_chk", "___strcp_D2A",
  "___strcpy_chk", "___strlcat_chk", "___strlcpy_chk", "___strncat_chk",
  "___strncpy_chk", "___strtodg", "___strtopdd", "___sum_D2A",
  "___svfscanf", "___swbuf", "___swhatbuf", "___swrite", "___swsetup",
  "___tens_D2A", "___tinytens_D2A", "___tolower", "___tolower_l",
  "___toupper", "___toupper_l", "___trailz_D2A", "___ulp_D2A",
  "___ungetc", "___ungetwc", "___vsnprintf_chk", "___vsprintf_chk",
  "___wcwidth", "___wcwidth_l", "__allocenvstate", "__atexit_receipt",
  "__c_locale", "__cleanup", "__closeutx", "__copyenv",
  "__cthread_init_routine", "__deallocenvstate", "__endutxent",
  "__flockfile_debug_stub", "__fseeko", "__ftello", "__fwalk",
  "__getenvp", "__getutxent", "__getutxid", "__getutxline",
  "__inet_aton_check", "__init_clock_port", "__int_to_time",
  "__libc_fork_child", "__libc_initializer", "__long_to_time",
  "__mkpath_np", "__mktemp", "__openutx", "__os_assert_log",
  "__os_assert_log_ctx", "__os_assumes_log", "__os_assumes_log_ctx",
  "__os_avoid_tail_call", "__os_crash", "__os_crash_callback",
  "__os_crash_fmt", "__os_debug_log", "__os_debug_log_error_str",
  "__putenvp", "__pututxline", "__rand48_add", "__rand48_mult",
  "__rand48_seed", "__readdir_unlocked", "__reclaim_telldir",
  "__seekdir", "__setenvp", "__setutxent", "__sigaction_nobind",
  "__sigintr", "__signal_nobind", "__sigvec_nobind", "__sread",
  "__sseek", "__subsystem_init", "__swrite", "__time32_to_time",
  "__time64_to_time", "__time_to_int", "__time_to_long",
  "__time_to_time32", "__time_to_time64", "__unsetenvp", "__utmpxname",
  "_a64l", "_abort", "_abort_report_np", "_abs", "_acl_add_flag_np",
  "_acl_add_perm", "_acl_calc_mask", "_acl_clear_flags_np",
  "_acl_clear_perms", "_acl_copy_entry", "_acl_copy_ext",
  "_acl_copy_ext_native", "_acl_copy_int", "_acl_copy_int_native",
  "_acl_create_entry", "_acl_create_entry_np", "_acl_delete_def_file",
  "_acl_delete_entry", "_acl_delete_fd_np", "_acl_delete_file_np",
  "_acl_delete_flag_np", "_acl_delete_link_np", "_acl_delete_perm",
  "_acl_dup", "_acl_free", "_acl_from_text", "_acl_get_entry",
  "_acl_get_fd", "_acl_get_fd_np", "_acl_get_file", "_acl_get_flag_np",
  "_acl_get_flagset_np", "_acl_get_link_np", "_acl_get_perm_np",
  "_acl_get_permset", "_acl_get_permset_mask_np", "_acl_get_qualifier",
  "_acl_get_tag_type", "_acl_init", "_acl_maximal_permset_mask_np",
  "_acl_set_fd", "_acl_set_fd_np", "_acl_set_file", "_acl_set_flagset_np",
  "_acl_set_link_np", "_acl_set_permset", "_acl_set_permset_mask_np",
  "_acl_set_qualifier", "_acl_set_tag_type", "_acl_size", "_acl_to_text",
  "_acl_valid", "_acl_valid_fd_np", "_acl_valid_file_np",
  "_acl_valid_link", "_addr2ascii", "_alarm", "_alphasort",
  "_arc4random", "_arc4random_addrandom", "_arc4random_buf",
  "_arc4random_stir", "_arc4random_uniform", "_ascii2addr", "_asctime",
  "_asctime_r", "_asprintf", "_asprintf_l", "_asxprintf",
  "_asxprintf_exec", "_atexit", "_atexit_b", "_atof", "_atof_l",
  "_atoi", "_atoi_l", "_atol", "_atol_l", "_atoll", "_atoll_l",
  "_backtrace", "_backtrace_from_fp", "_backtrace_image_offsets",
  "_backtrace_symbols", "_backtrace_symbols_fd", "_basename",
  "_basename_r", "_bcopy", "_brk", "_bsd_signal", "_bsearch",
  "_bsearch_b", "_btowc", "_btowc_l", "_catclose", "_catgets",
  "_catopen", "_cfgetispeed", "_cfgetospeed", "_cfmakeraw",
  "_cfsetispeed", "_cfsetospeed", "_cfsetspeed", "_cgetcap",
  "_cgetclose", "_cgetent", "_cgetfirst", "_cgetmatch", "_cgetnext",
  "_cgetnum", "_cgetset", "_cgetstr", "_cgetustr", "_chmodx_np",
  "_clearerr", "_clearerr_unlocked", "_clock", "_clock_getres",
  "_clock_gettime", "_clock_gettime_nsec_np", "_clock_port",
  "_clock_sem", "_clock_settime", "_closedir", "_compat_mode",
  "_confstr", "_copy_printf_domain", "_creat", "_crypt", "_ctermid",
  "_ctermid_r", "_ctime", "_ctime_r", "_daemon", "_daylight",
  "_dbm_clearerr", "_dbm_close", "_dbm_delete", "_dbm_dirfno",
  "_dbm_error", "_dbm_fetch", "_dbm_firstkey", "_dbm_nextkey",
  "_dbm_open", "_dbm_store", "_dbopen", "_devname", "_devname_r",
  "_difftime", "_digittoint", "_digittoint_l", "_dirfd", "_dirname",
  "_dirname_r", "_div", "_dprintf", "_dprintf_l", "_drand48",
  "_duplocale", "_dxprintf", "_dxprintf_exec", "_ecvt", "_encrypt",
  "_endttyent", "_endusershell", "_endutxent", "_endutxent_wtmp",
  "_erand48", "_err", "_err_set_exit", "_err_set_exit_b",
  "_err_set_file", "_errc", "_errx", "_execl", "_execle", "_execlp",
  "_execv", "_execvP", "_execvp", "_exit", "_f_prealloc", "_fchmodx_np",
  "_fclose", "_fcvt", "_fdopen", "_fdopendir", "_feof", "_feof_unlocked",
  "_ferror", "_ferror_unlocked", "_fflagstostr", "_fflush", "_fgetc",
  "_fgetln", "_fgetpos", "_fgetrune", "_fgets", "_fgetwc", "_fgetwc_l",
  "_fgetwln", "_fgetwln_l", "_fgetws", "_fgetws_l", "_fileno",
  "_fileno_unlocked", "_filesec_dup", "_filesec_free",
  "_filesec_get_property", "_filesec_init", "_filesec_query_property",
  "_filesec_set_property", "_filesec_unset_property", "_flockfile",
  "_fmemopen", "_fmtcheck", "_fmtmsg", "_fnmatch", "_fopen", "_fork",
  "_forkpty", "_fparseln", "_fprintf", "_fprintf_l", "_fpurge",
  "_fputc", "_fputrune", "_fputs", "_fputwc", "_fputwc_l", "_fputws",
  "_fputws_l", "_fread", "_free_printf_comp", "_free_printf_domain",
  "_freelocale", "_freopen", "_fscanf", "_fscanf_l", "_fseek",
  "_fseeko", "_fsetpos", "_fstatvfs", "_fstatx_np", "_fsync_volume_np",
  "_ftell", "_ftello", "_ftime", "_ftok", "_ftrylockfile",
  "_fts_children", "_fts_close", "_fts_open", "_fts_open_b",
  "_fts_read", "_fts_set", "_ftw", "_fungetrune", "_funlockfile",
  "_funopen", "_fwide", "_fwprintf", "_fwprintf_l", "_fwrite",
  "_fwscanf", "_fwscanf_l", "_fxprintf", "_fxprintf_exec", "_gcvt",
  "_getbsize", "_getc", "_getc_unlocked", "_getchar", "_getchar_unlocked",
  "_getcwd", "_getdate", "_getdate_err", "_getdelim", "_getdiskbyname",
  "_getenv", "_gethostid", "_gethostname", "_getipv4sourcefilter",
  "_getlastlogx", "_getlastlogxbyname", "_getline", "_getloadavg",
  "_getlogin", "_getlogin_r", "_getmntinfo", "_getmntinfo_r_np",
  "_getmode", "_getopt", "_getopt_long", "_getopt_long_only",
  "_getpagesize", "_getpass", "_getpeereid", "_getprogname", "_gets",
  "_getsourcefilter", "_getsubopt", "_gettimeofday", "_getttyent",
  "_getttynam", "_getusershell", "_getutmp", "_getutmpx", "_getutxent",
  "_getutxent_wtmp", "_getutxid", "_getutxline", "_getvfsbyname",
  "_getw", "_getwc", "_getwc_l", "_getwchar", "_getwchar_l", "_getwd",
  "_glob", "_glob_b", "_globfree", "_gmtime", "_gmtime_r", "_grantpt",
  "_hash_create", "_hash_destroy", "_hash_purge", "_hash_search",
  "_hash_stats", "_hash_traverse", "_hcreate", "_hdestroy",
  "_heapsort", "_heapsort_b", "_hsearch", "_imaxabs", "_imaxdiv",
  "_inet_addr", "_inet_aton", "_inet_lnaof", "_inet_makeaddr",
  "_inet_net_ntop", "_inet_net_pton", "_inet_neta", "_inet_netof",
  "_inet_network", "_inet_nsap_addr", "_inet_nsap_ntoa", "_inet_ntoa",
  "_inet_ntop", "_inet_ntop4", "_inet_ntop6", "_inet_pton",
  "_initstate", "_insque", "_isalnum", "_isalnum_l", "_isalpha",
  "_isalpha_l", "_isascii", "_isatty", "_isblank", "_isblank_l",
  "_iscntrl", "_iscntrl_l", "_isdigit", "_isdigit_l", "_isgraph",
  "_isgraph_l", "_ishexnumber", "_ishexnumber_l", "_isideogram",
  "_isideogram_l", "_islower", "_islower_l", "_isnumber", "_isnumber_l",
  "_isphonogram", "_isphonogram_l", "_isprint", "_isprint_l",
  "_ispunct", "_ispunct_l", "_isrune", "_isrune_l", "_isspace",
  "_isspace_l", "_isspecial", "_isspecial_l", "_isupper", "_isupper_l",
  "_iswalnum", "_iswalnum_l", "_iswalpha", "_iswalpha_l", "_iswascii",
  "_iswblank", "_iswblank_l", "_iswcntrl", "_iswcntrl_l", "_iswctype",
  "_iswctype_l", "_iswdigit", "_iswdigit_l", "_iswgraph", "_iswgraph_l",
  "_iswhexnumber", "_iswhexnumber_l", "_iswideogram", "_iswideogram_l",
  "_iswlower", "_iswlower_l", "_iswnumber", "_iswnumber_l",
  "_iswphonogram", "_iswphonogram_l", "_iswprint", "_iswprint_l",
  "_iswpunct", "_iswpunct_l", "_iswrune", "_iswrune_l", "_iswspace",
  "_iswspace_l", "_iswspecial", "_iswspecial_l", "_iswupper",
  "_iswupper_l", "_iswxdigit", "_iswxdigit_l", "_isxdigit",
  "_isxdigit_l", "_jrand48", "_kOSThermalNotificationPressureLevelName",
  "_killpg", "_l64a", "_labs", "_lchflags", "_lchmod", "_lcong48",
  "_ldiv", "_lfind", "_link_addr", "_link_ntoa", "_llabs", "_lldiv",
  "_localeconv", "_localeconv_l", "_localtime", "_localtime_r",
  "_lockf", "_login", "_login_tty", "_logout", "_logwtmp", "_lrand48",
  "_lsearch", "_lstatx_np", "_lutimes", "_mblen", "_mblen_l",
  "_mbmb", "_mbrlen", "_mbrlen_l", "_mbrrune", "_mbrtowc", "_mbrtowc_l",
  "_mbrune", "_mbsinit", "_mbsinit_l", "_mbsnrtowcs", "_mbsnrtowcs_l",
  "_mbsrtowcs", "_mbsrtowcs_l", "_mbstowcs", "_mbstowcs_l", "_mbtowc",
  "_mbtowc_l", "_memmem", "_memset_s", "_mergesort", "_mergesort_b",
  "_mkdirx_np", "_mkdtemp", "_mkdtempat_np", "_mkfifox_np",
  "_mkostemp", "_mkostemps", "_mkostempsat_np", "_mkpath_np",
  "_mkpathat_np", "_mkstemp", "_mkstemp_dprotected_np", "_mkstemps",
  "_mkstempsat_np", "_mktemp", "_mktime", "_monaddition", "_moncontrol",
  "_moncount", "_moninit", "_monitor", "_monoutput", "_monreset",
  "_monstartup", "_mpool_close", "_mpool_filter", "_mpool_get",
  "_mpool_new", "_mpool_open", "_mpool_put", "_mpool_sync", "_mrand48",
  "_nanosleep", "_new_printf_comp", "_new_printf_domain", "_newlocale",
  "_nextwctype", "_nextwctype_l", "_nftw", "_nice", "_nl_langinfo",
  "_nl_langinfo_l", "_nrand48", "_nvis", "_off32", "_off64",
  "_offtime", "_open_memstream", "_open_with_subsystem",
  "_open_wmemstream", "_opendev", "_opendir", "_openpty", "_openx_np",
  "_optarg", "_opterr", "_optind", "_optopt", "_optreset", "_pause",
  "_pclose", "_perror", "_popen", "_posix2time", "_posix_openpt",
  "_posix_spawnp", "_printf", "_printf_l", "_psignal", "_psort",
  "_psort_b", "_psort_r", "_ptsname", "_ptsname_r", "_putc",
  "_putc_unlocked", "_putchar", "_putchar_unlocked", "_putenv",
  "_puts", "_pututxline", "_putw", "_putwc", "_putwc_l", "_putwchar",
  "_putwchar_l", "_qsort", "_qsort_b", "_qsort_r", "_querylocale",
  "_radixsort", "_raise", "_rand", "_rand_r", "_random", "_rb_tree_count",
  "_rb_tree_find_node", "_rb_tree_find_node_geq", "_rb_tree_find_node_leq",
  "_rb_tree_init", "_rb_tree_insert_node", "_rb_tree_iterate",
  "_rb_tree_remove_node", "_readdir", "_readdir_r", "_readpassphrase",
  "_reallocf", "_realpath", "_recv", "_regcomp", "_regcomp_l",
  "_regerror", "_regexec", "_regfree", "_register_printf_domain_function",
  "_register_printf_domain_render_std", "_regncomp", "_regncomp_l",
  "_regnexec", "_regwcomp", "_regwcomp_l", "_regwexec", "_regwncomp",
  "_regwncomp_l", "_regwnexec", "_remove", "_remque", "_rewind",
  "_rewinddir", "_rindex", "_rpmatch", "_sbrk", "_scandir",
  "_scandir_b", "_scanf", "_scanf_l", "_seed48", "_seekdir", "_send",
  "_setbuf", "_setbuffer", "_setenv", "_sethostid", "_sethostname",
  "_setinvalidrune", "_setipv4sourcefilter", "_setkey", "_setlinebuf",
  "_setlocale", "_setlogin", "_setmode", "_setpgrp", "_setprogname",
  "_setrgid", "_setruid", "_setrunelocale", "_setsourcefilter",
  "_setstate", "_settimeofday", "_setttyent", "_setusershell",
  "_setutxent", "_setutxent_wtmp", "_setvbuf", "_sigaction",
  "_sigaddset", "_sigaltstack", "_sigblock", "_sigdelset",
  "_sigemptyset", "_sigfillset", "_sighold", "_sigignore",
  "_siginterrupt", "_sigismember", "_signal", "_sigpause", "_sigrelse",
  "_sigset", "_sigsetmask", "_sigvec", "_skip", "_sl_add", "_sl_find",
  "_sl_free", "_sl_init", "_sleep", "_snprintf", "_snprintf_l",
  "_snvis", "_sockatmark", "_sprintf", "_sprintf_l", "_sradixsort",
  "_srand", "_srand48", "_sranddev", "_srandom", "_srandomdev",
  "_sscanf", "_sscanf_l", "_stat_with_subsystem", "_statvfs",
  "_statx_np", "_stpcpy", "_stpncpy", "_strcasecmp", "_strcasecmp_l",
  "_strcasestr", "_strcasestr_l", "_strcat", "_strcoll", "_strcoll_l",
  "_strcspn", "_strdup", "_strenvisx", "_strerror", "_strerror_r",
  "_strfmon", "_strfmon_l", "_strftime", "_strftime_l", "_strmode",
  "_strncasecmp", "_strncasecmp_l", "_strncat", "_strndup", "_strnstr",
  "_strnunvis", "_strnunvisx", "_strnvis", "_strnvisx", "_strpbrk",
  "_strptime", "_strptime_l", "_strrchr", "_strsenvisx", "_strsep",
  "_strsignal", "_strsignal_r", "_strsnvis", "_strsnvisx", "_strspn",
  "_strsvis", "_strsvisx", "_strtod", "_strtod_l", "_strtof",
  "_strtof_l", "_strtofflags", "_strtoimax", "_strtoimax_l",
  "_strtok", "_strtok_r", "_strtol", "_strtol_l", "_strtold",
  "_strtold_l", "_strtoll", "_strtoll_l", "_strtonum", "_strtoq",
  "_strtoq_l", "_strtoul", "_strtoul_l", "_strtoull", "_strtoull_l",
  "_strtoumax", "_strtoumax_l", "_strtouq", "_strtouq_l", "_strunvis",
  "_strunvisx", "_strvis", "_strvisx", "_strxfrm", "_strxfrm_l",
  "_suboptarg", "_svis", "_swab", "_swprintf", "_swprintf_l",
  "_swscanf", "_swscanf_l", "_sxprintf", "_sxprintf_exec",
  "_sync_volume_np", "_sys_errlist", "_sys_nerr", "_sys_siglist",
  "_sys_signame", "_sysconf", "_sysctl", "_sysctlbyname",
  "_sysctlnametomib", "_system", "_tcdrain", "_tcflow", "_tcflush",
  "_tcgetattr", "_tcgetpgrp", "_tcgetsid", "_tcsendbreak", "_tcsetattr",
  "_tcsetpgrp", "_tdelete", "_telldir", "_tempnam", "_tfind",
  "_thread_stack_pcs", "_time", "_time2posix", "_timegm", "_timelocal",
  "_timeoff", "_times", "_timespec_get", "_timezone", "_timingsafe_bcmp",
  "_tmpfile", "_tmpnam", "_toascii", "_tolower", "_tolower_l",
  "_toupper", "_toupper_l", "_towctrans", "_towctrans_l", "_towlower",
  "_towlower_l", "_towupper", "_towupper_l", "_tre_ast_new_catenation",
  "_tre_ast_new_iter", "_tre_ast_new_literal", "_tre_ast_new_node",
  "_tre_ast_new_union", "_tre_compile", "_tre_fill_pmatch",
  "_tre_free", "_tre_mem_alloc_impl", "_tre_mem_destroy",
  "_tre_mem_new_impl", "_tre_parse", "_tre_stack_destroy",
  "_tre_stack_new", "_tre_stack_num_objects", "_tre_tnfa_run_backtrack",
  "_tre_tnfa_run_parallel", "_tsearch", "_ttyname", "_ttyname_r",
  "_ttyslot", "_twalk", "_tzname", "_tzset", "_tzsetwall", "_ualarm",
  "_ulimit", "_umaskx_np", "_uname", "_ungetc", "_ungetwc",
  "_ungetwc_l", "_unlockpt", "_unsetenv", "_unvis", "_uselocale",
  "_usleep", "_utime", "_utmpxname", "_uuid_clear", "_uuid_compare",
  "_uuid_copy", "_uuid_generate", "_uuid_generate_random",
  "_uuid_generate_time", "_uuid_is_null", "_uuid_pack", "_uuid_parse",
  "_uuid_unpack", "_uuid_unparse", "_uuid_unparse_lower",
  "_uuid_unparse_upper", "_vasprintf", "_vasprintf_l", "_vasxprintf",
  "_vasxprintf_exec", "_vdprintf", "_vdprintf_l", "_vdxprintf",
  "_vdxprintf_exec", "_verr", "_verrc", "_verrx", "_vfprintf",
  "_vfprintf_l", "_vfscanf", "_vfscanf_l", "_vfwprintf", "_vfwprintf_l",
  "_vfwscanf", "_vfwscanf_l", "_vfxprintf", "_vfxprintf_exec",
  "_vis", "_vprintf", "_vprintf_l", "_vscanf", "_vscanf_l",
  "_vsnprintf", "_vsnprintf_l", "_vsprintf", "_vsprintf_l", "_vsscanf",
  "_vsscanf_l", "_vswprintf", "_vswprintf_l", "_vswscanf",
  "_vswscanf_l", "_vsxprintf", "_vsxprintf_exec", "_vwarn", "_vwarnc",
  "_vwarnx", "_vwprintf", "_vwprintf_l", "_vwscanf", "_vwscanf_l",
  "_vxprintf", "_vxprintf_exec", "_wait", "_wait3", "_waitpid",
  "_warn", "_warnc", "_warnx", "_wcpcpy", "_wcpncpy", "_wcrtomb",
  "_wcrtomb_l", "_wcscasecmp", "_wcscasecmp_l", "_wcscat", "_wcschr",
  "_wcscmp", "_wcscoll", "_wcscoll_l", "_wcscpy", "_wcscspn",
  "_wcsdup", "_wcsftime", "_wcsftime_l", "_wcslcat", "_wcslcpy",
  "_wcslen", "_wcsncasecmp", "_wcsncasecmp_l", "_wcsncat", "_wcsncmp",
  "_wcsncpy", "_wcsnlen", "_wcsnrtombs", "_wcsnrtombs_l", "_wcspbrk",
  "_wcsrchr", "_wcsrtombs", "_wcsrtombs_l", "_wcsspn", "_wcsstr",
  "_wcstod", "_wcstod_l", "_wcstof", "_wcstof_l", "_wcstoimax",
  "_wcstoimax_l", "_wcstok", "_wcstol", "_wcstol_l", "_wcstold",
  "_wcstold_l", "_wcstoll", "_wcstoll_l", "_wcstombs", "_wcstombs_l",
  "_wcstoul", "_wcstoul_l", "_wcstoull", "_wcstoull_l", "_wcstoumax",
  "_wcstoumax_l", "_wcswidth", "_wcswidth_l", "_wcsxfrm", "_wcsxfrm_l",
  "_wctob", "_wctob_l", "_wctomb", "_wctomb_l", "_wctrans",
  "_wctrans_l", "_wctype", "_wctype_l", "_wcwidth", "_wcwidth_l",
  "_wmemchr", "_wmemcmp", "_wmemcpy", "_wmemmove", "_wmemset",
  "_wordexp", "_wordfree", "_wprintf", "_wprintf_l", "_wscanf",
  "_wscanf_l", "_wtmpxname", "_xprintf", "_xprintf_exec"
)

def print_here_head(name):
  print("""\
(tee %s.s |llvm-mc -filetype=obj -triple %s -o %s.o) <<END_OF_FILE &""" % (name, triple, name))

def print_here_tail():
  print("""\
END_OF_FILE
""")

def print_function_head(p2align, name):
  if args.os == "macos":
      print("""\
    .section __TEXT,__text,regular,pure_instructions
    .p2align %d, 0x90
    .globl _%s
_%s:""" % (p2align, name, name))
  elif args.os == "windows":
      print("""\
    .text
    .def %s;
    .scl 2;
    .type 32;
    .endef
    .globl %s
    .p2align %d
%s:""" % (name, name, p2align, name))
  elif args.os == "linux":
      print("""\
    .text
    .p2align %d
    .globl %s
%s:""" % (p2align, name, name))

def print_function(addr, size, addrs):
  name = "x%08x" % addr
  calls = random.randint(0, size>>12)
  print_here_head(name)
  print("""\
### %s size=%x calls=%x""" % (name, size, calls))
  print_function_head(4, name)
  for i in range(calls):
      print("    bl %sx%08x\n    .p2align 4" %
            ("_" if args.os == "macos" else "",
             addrs[random.randint(0, len(addrs)-1)]))
      if args.os == "macos":
        print("    bl %s\n    .p2align 4" %
              (libSystem_calls[random.randint(0, len(libSystem_calls)-1)]))
  fill = size - 4 * (calls + 1)
  assert fill > 0
  print("""\
    .fill 0x%x
    ret""" % (fill))
  print_here_tail()

def random_seed():
  """Generate a seed that can easily be passsed back in via --seed=STRING"""
  return ''.join(random.choice(string.ascii_lowercase) for i in range(10))

def generate_sizes(base, megabytes):
  total = 0
  while total < megabytes:
      size = random.randint(0x100, 0x10000) * 0x10
      yield size
      total += size

def generate_addrs(addr, sizes):
  i = 0
  while i < len(sizes):
      yield addr
      addr += sizes[i]
      i += 1

def main():
  parser = argparse.ArgumentParser(
    description=__doc__,
    epilog="""\
WRITEME
""")
  parser.add_argument('--seed', type=str, default=random_seed(),
                      help='Seed the random number generator')
  parser.add_argument('--size', type=int, default=None,
                      help='Total text size to generate, in megabytes')
  parser.add_argument('--os', type=str, default="macos",
                      help='Target OS: macos, windows, or linux')
  global args
  args = parser.parse_args()
  triples = {
      "macos": "arm64-apple-macos",
      "linux": "aarch64-pc-linux",
      "windows": "aarch64-pc-windows"
  }
  global triple
  triple = triples.get(args.os)

  print("""\
### seed=%s triple=%s
""" % (args.seed, triple))

  random.seed(args.seed)

  base = 0x4010
  megabytes = (int(args.size) if args.size else 512) * 1024 * 1024
  sizes = [size for size in generate_sizes(base, megabytes)]
  addrs = [addr for addr in generate_addrs(base, sizes)]

  for i in range(len(addrs)):
      print_function(addrs[i], sizes[i], addrs)

  print_here_head("main")
  print("""\
### _x%08x
""" % (addrs[-1] + sizes[-1]))
  print_function_head(14 if args.os == "macos" else 4, "main")
  print("    ret")
  print_here_tail()
  print("wait")


if __name__ == '__main__':
  main()
